Theoretically Efficient Parallel Graph Algorithms
Can Be Fast and Scalable
Abstract.
There has been significant recent interest in parallel graph processing due to the need to quickly analyze the large graphs available today. Many graph codes have been designed for distributed memory or external memory. However, today even the largest publicly-available real-world graph (the Hyperlink Web graph with over 3.5 billion vertices and 128 billion edges) can fit in the memory of a single commodity multicore server. Nevertheless, most experimental work in the literature report results on much smaller graphs, and the ones for the Hyperlink graph use distributed or external memory. Therefore, it is natural to ask whether we can efficiently solve a broad class of graph problems on this graph in memory.
This paper shows that theoretically-efficient parallel graph algorithms can scale to the largest publicly-available graphs using a single machine with a terabyte of RAM, processing them in minutes. We give implementations of theoretically-efficient parallel algorithms for 20 important graph problems. We also present the optimizations and techniques that we used in our implementations, which were crucial in enabling us to process these large graphs quickly. We show that the running times of our implementations outperform existing state-of-the-art implementations on the largest real-world graphs. For many of the problems that we consider, this is the first time they have been solved on graphs at this scale. We have made the implementations developed in this work publicly-available as the Graph-Based Benchmark Suite (GBBS).
1. Introduction
| Problem | (1) | (72h) | (SU) | Alg. | Model | Work | Depth |
| Breadth-First Search (BFS) | 576 | 8.44 | 68 | – | |||
| Integral-Weight SSSP (weighted BFS) | 3770 | 58.1 | 64 | (42) | expected | w.h.p.† | |
| General-Weight SSSP (Bellman-Ford) | 4010 | 59.4 | 67 | (38) | |||
| Single-Source Widest Path (Bellman-Ford) | 3210 | 48.4 | 66 | (38) | |||
| Single-Source Betweenness Centrality (BC) | 2260 | 37.1 | 60 | (30) | |||
| -Spanner | 2390 | 36.5 | 65 | (89) | w.h.p. | ||
| Low-Diameter Decomposition (LDD) | 980 | 16.6 | 59 | (90) | w.h.p. | ||
| Connectivity | 1640 | 25.0 | 65 | (117) | expected | w.h.p. | |
| Spanning Forest | 2420 | 35.8 | 67 | (117) | expected | w.h.p. | |
| Biconnectivity | 9860 | 165 | 59 | (125) | expected | w.h.p. | |
| Strongly Connected Components (SCC)* | 8130 | 185 | 43 | (23) | expected | w.h.p. | |
| Minimum Spanning Forest (MSF) | 9520 | 187 | 50 | (130) | |||
| Maximal Independent Set (MIS) | 2190 | 32.2 | 68 | (22) | expected | w.h.p. | |
| Maximal Matching (MM) | 7150 | 108 | 66 | (22) | expected | w.h.p. | |
| Graph Coloring | 8920 | 158 | 56 | (59) | |||
| Approximate Set Cover | 5320 | 90.4 | 58 | (25) | expected | w.h.p. | |
| -core | 8515 | 184 | 46 | (42) | expected | w.h.p. | |
| Approximate Densest Subgraph | 3780 | 51.4 | 73 | (13) | |||
| Triangle Counting (TC) | — | 1168 | — | (119) | – | ||
| PageRank Iteration | 973 | 13.1 | 74 | (31) |
Today, the largest publicly-available graph, the Hyperlink Web graph, consists of over 3.5 billion vertices and 128 billion edges (86). This graph presents a significant computational challenge for both distributed and shared memory systems. Indeed, very few algorithms have been applied to this graph, and those that have often take hours to run (39, 79, 67), with the fastest times requiring between 1–6 minutes using a supercomputer (122, 123). In this paper, we show that a wide range of fundamental graph problems can be solved quickly on this graph, often in minutes, on a single commodity shared-memory machine with a terabyte of RAM.11 1 These machines are roughly the size of a workstation and can be easily rented in the cloud (e.g., on Amazon EC2). For example, our -core implementation takes under 3.5 minutes on 72 cores, whereas Slota et al. (123) report a running time of about 6 minutes for approximate -core on a supercomputer with over 8000 cores. They also report that they can identify the largest connected component on this graph in 63 seconds, whereas we can identify all connected components in 25 seconds. Another recent result by Stergiou et al. (124) solves connectivity on the Hyperlink 2012 graph in 341 seconds on a 1000 node cluster with 12000 cores and 128TB of RAM. Compared to this result, our implementation is 13.6x faster on a system with 128x less memory and 166x fewer cores. However, we note that they are able to process a significantly larger private graph that we would not be able to fit into our memory footprint. A more complete comparison between our work and existing work, including both distributed and disk-based systems (39, 79, 67, 41, 60), is given in Section 7.
Importantly, all of our implementations have strong theoretical bounds on their work and depth. There are several reasons that algorithms with good theoretical guarantees are desirable. For one, they are robust as even adversarially-chosen inputs will not cause them to perform extremely poorly. Furthermore, they can be designed on pen-and-paper by exploiting properties of the problem instead of tailoring solutions to the particular dataset at hand. Theoretical guarantees also make it likely that the algorithm will continue to perform well even if the underlying data changes. Finally, careful implementations of algorithms that are nearly work-efficient can perform much less work in practice than work-inefficient algorithms. This reduction in work often translates to faster running times on the same number of cores (42). We note that most running times that have been reported in the literature on the Hyperlink Web graph use parallel algorithms that are not theoretically-efficient.
In this paper, we present implementations of parallel algorithms with strong theoretical bounds on their work and depth for connectivity, biconnectivity, strongly connected components, low-diameter decomposition, graph spanners, maximal independent set, maximal matching, graph coloring, breadth-first search, single-source shortest paths, widest (bottleneck) path, betweenness centrality, PageRank, spanning forest, minimum spanning forest, -core decomposition, approximate set cover, approximate densest subgraph, and triangle counting. We describe the techniques used to achieve good performance on graphs with billions of vertices and hundreds of billions of edges and share experimental results for the Hyperlink 2012 and Hyperlink 2014 Web crawls, the largest and second largest publicly-available graphs, as well as several smaller real-world graphs at various scales. Some of the algorithms we describe are based on previous results from Ligra, Ligra+, and Julienne (114, 118, 42), and other papers on efficient parallel graph algorithms (22, 59, 119). However, most existing implementations were changed significantly in order to be more memory efficient. Several algorithm implementations for problems like strongly connected components, minimum spanning forest, and biconnectivity are new, and required implementation techniques to scale that we believe are of independent interest. We also had to extend the compressed representation from Ligra+ (118) to ensure that our graph primitives for mapping, filtering, reducing and packing the neighbors of a vertex were theoretically-efficient. We note that using compression techniques is crucial for representing the symmetrized Hyperlink 2012 graph in 1TB of RAM, as storing this graph in an uncompressed format would require over 900GB to store the edges alone, whereas the graph requires 330GB in our compressed format (less than 1.5 bytes per edge). We show the running times of our algorithms on the Hyperlink 2012 graph as well as their work and depth bounds in Table 1. To make it easy to build upon or compare to our work in the future, we describe a benchmark suite containing our problems with clear I/O specifications, which we have made publicly-available.22 2 https://github.com/ldhulipala/gbbs
We present an experimental evaluation of all of our implementations, and in almost all cases, the numbers we report are faster than any previous performance numbers for any machines, even much larger supercomputers. We are also able to apply our algorithms to the largest publicly-available graph, in many cases for the first time in the literature, using a reasonably modest machine. Most importantly, our implementations are based on reasonably simple algorithms with strong bounds on their work and depth. We believe that our implementations are likely to scale to larger graphs and lead to efficient algorithms for related problems.
2. Related Work
Parallel Graph Algorithms. Parallel graph algorithms have received significant attention since the start of parallel computing, and many elegant algorithms with good theoretical bounds have been developed over the decades (e.g., (113, 71, 78, 4, 125, 91, 104, 66, 34, 100, 90, 46, 18, 88)). A major goal in parallel graph algorithm design is to find work-efficient algorithms with polylogarithmic depth. While many suspect that work-efficient algorithms may not exist for all parallelizable graph problems, as inefficiency may be inevitable for problems that depend on transitive closure, many problems that are of practical importance do admit work-efficient algorithms (70). For these problems, which include connectivity, biconnectivity, minimum spanning forest, maximal independent set, maximal matching, and triangle counting, giving theoretically-efficient implementations that are simple and practical is important, as the amount of parallelism available on modern systems is still modest enough that reducing the amount of work done is critical for achieving good performance. Aside from intellectual curiosity, investigating whether theoretically-efficient graph algorithms also perform well in practice is important, as theoretically-efficient algorithms are less vulnerable to adversarial inputs than ad-hoc algorithms that happen to work well in practice.
Unfortunately, some problems that are not known to admit work-efficient parallel algorithms due to the transitive-closure bottleneck (70), such as strongly connected components (SCC) and single-source shortest paths (SSSP) are still important in practice. One method for circumventing the bottleneck is to give work-efficient algorithms for these problems that run in depth proportional to the diameter of the graph—as real-world graphs have low diameter, and theoretical models of real-world graphs predict a logarithmic diameter, these algorithms offer theoretical guarantees in practice (110, 23). Other problems, like -core are -complete (6), which rules out polylogarithmic-depth algorithms for them unless (55). However, even -core admits an algorithm with strong theoretical guarantees that is efficient in practice (42).
Parallel Graph Processing Frameworks. Motivated by the need to process very large graphs, there have been many graph processing frameworks developed in the literature (e.g., (81, 53, 77, 93, 114) among many others). We refer the reader to (84, 128) for surveys of existing frameworks. Several recent graph processing systems evaluate the scalability of their implementations by solving problems on massive graphs (122, 39, 79, 42, 67, 124). All of these systems report running times either on the Hyperlink 2012 graph or Hyperlink 2014 graphs, two web crawls released by the WebDataCommons that are the largest and second largest publicly-available graphs respectively. We describe these recent systems and give a detailed comparison of how our implementations perform compare to their codes in Section 7.
Benchmarking Parallel Graph Algorithms. There are a surprising number of existing benchmarks of parallel graph algorithms. SSCA (10), an early graph processing benchmark, specifies four graph kernels, which include generating graphs in adjacency list format, subgraph extraction, and graph clustering. The Problem Based Benchmark Suite (PBBS) (116) is a general benchmark of parallel algorithms that includes 6 problems on graphs (BFS, spanning forest, minimum spanning forest, maximal independent set, maximal matching, and graph separators). The PBBS benchmarks are problem-based in that they are defined only in terms of the input and output without any specification of the algorithm used to solve the problem. We follow the style of PBBS in this paper of defining the input and output requirements for each problem. The Graph Algorithm Platform (GAP) Benchmark Suite (15) specifies 6 kernels for BFS, SSSP, PageRank, connectivity, betweenness centrality, and triangle counting.
Several recent benchmarks characterize the architectural properties of parallel graph algorithms. GraphBIG (92) describes 12 applications, including several problems that we consider, like -core and graph coloring (using the Jones-Plassmann algorithm), but also problems like depth-first search, which are difficult to parallelize, as well as dynamic graph operations. CRONO (3) implements 10 graph algorithms, including all-pairs shortest paths, exact betweenness centrality, traveling salesman, and depth-first search. LDBC (63) is an industry-driven benchmark that selects 6 algorithms that are considered representative of graph processing including BFS, and several algorithms based on label propagation.
Unfortunately, all of the existing graph algorithm benchmarks we are aware of restrict their evaluation to small graphs, often on the order of tens or hundreds of millions of edges, with the largest graphs in the benchmarks having about two billion edges. As real-world graphs are frequently several orders of magnitude larger than this, evaluation on such small graphs makes it hard to judge whether the algorithms or results from a benchmark scale to terabyte-scale graphs.
3. Preliminaries
Graph Notation. We denote an unweighted graph by , where is the set of vertices and is the set of edges in the graph. A weighted graph is denoted by , where is a function which maps an edge to a real value (its weight). The number of vertices in a graph is , and the number of edges is . Vertices are assumed to be indexed from to . For undirected graphs we use to denote the neighbors of vertex and to denote its degree. For directed graphs, we use and to denote the in and out-neighbors of a vertex . We use to refer to the diameter of the graph, or the longest shortest path distance between any vertex and any vertex reachable from . Given an undirected graph the density of a set , or , is equal to where are the edges in the induced subgraph on . is used to denote the maximum degree of the graph. We assume that there are no self-edges or duplicate edges in the graph. We refer to graphs stored as a list of edges as being stored in the edgelist format and the compressed-sparse column and compressed-sparse row formats as CSC and CSR respectively.
Atomic Primitives. We use three common atomic primitives in our algorithms: test-and-set (TS), fetch-and-add (FA), and priority-write (PW). A checks if is , and if so atomically sets it to and returns ; otherwise it returns . A atomically returns the current value of and then increments . A atomically compares with the current value of using the priority function , and if has higher priority than the value of according to it sets to and returns ; otherwise it returns .
Model. In the analysis of algorithms we use the following work-depth model, which is closely related to the but better models current machines and programming paradigms that are asynchronous and allow dynamic forking. We can simulate the model on the equipped with the same operations with an additional factor in the depth due to load-balancing. Furthermore, a algorithm using processors and time can be simulated in our model with work and depth.
The Threaded Random-Access Machine () (20) consists of a set of threads that share an unbounded memory. Each thread is basically equivalent to a Random Access Machine—it works on a program stored in memory, has a constant number of registers, and has standard instructions (including an end to finish the computation). The extends the with a fork instruction that takes a positive integer and forks new child threads. Each child thread receives a unique integer in the range in its first register and otherwise has the identical state as the parent, which has a in that register. They all start by running the next instruction. When a thread performs a fork, it is suspended until all the children terminate (execute an end instruction). A computation starts with a single root thread and finishes when that root thread ends. This model supports what is often referred to as nested parallelism. If the root thread never does a fork, it is a standard sequential program.
A computation can be viewed as a series-parallel DAG in which each instruction is a vertex, sequential instructions are composed in series, and the forked subthreads are composed in parallel. The work of a computation is the number of vertices and the depth is the length of the longest path in the DAG. We augment the model with three atomic instructions that are used by our algorithms: test-and-set (), fetch-and-add (), and priority-write () and discuss our model with these operations as the , , and variants of the . As is standard with the model, we assume that the memory locations and registers have at most bits, where is the total size of the memory used. More details about the model can be found in (20).
Parallel Primitives. The following parallel procedures are used throughout the paper. Scan takes as input an array of length , an associative binary operator , and an identity element such that for any , and returns the array as well as the overall sum, . Scan can be done in work and depth (assuming takes work) (66). Reduce takes an array and a binary associative function and returns the sum of the elements in with respect to . Filter takes an array and a predicate and returns a new array containing for which is true, in the same order as in . Reduce and filter can both be done in work and depth (assuming takes work).
Ligra, Ligra+, and Julienne. We make use of the Ligra, Ligra+, and Julienne frameworks for shared-memory graph processing in this paper and review components from these frameworks here (114, 118, 42). Ligra provides data structures for representing a graph , vertexSubsets (subsets of the vertices). We make use of the edgeMap function provided by Ligra, which we use for mapping over edges.
edgeMap takes as input a graph , a vertexSubset , and two boolean functions and . edgeMap applies to such that and (call this subset of edges ), and returns a vertexSubset where if and only if and . can side-effect data structures associated with the vertices. edgeMap runs in work and depth assuming and take work. edgeMap either applies a sparse or dense method based on the number of edges incident to the current frontier. Both methods run in work and depth. We note that in our experiments we use an optimized version of the dense method which examines in-edges sequentially and stops once returns . This optimization lets us potentially examine significantly fewer edges than the depth version, but at the cost of depth.
4. Benchmark
In this section we describe I/O specifications of our benchmark. We discuss related work and present the theoretically-efficient algorithm implemented for each problem in Section 5. We mark implementations based on prior work with a .
4.1. Shortest Path Problems
4.2. Connectivity Problems
4.3. Covering Problems
4.4. Substructure Problems
4.5. Eigenvector Problems
5. Algorithms
In this section we discuss related work and give self-contained descriptions of all of the theoretically efficient algorithms implemented in our benchmark. We also include implementation details for algorithms implemented in prior papers that we did not significantly change. The pseudocode for many of the algorithms make use of the edgeMap and vertexMap primitives, as well as test-and-set, fetch-and-add and priority-write. All of these primitives are defined in Section 3. We cite the original papers that our algorithms are based on in Table 1.
5.1. Shortest Path Problems
Although work-efficient polylogarithmic-depth algorithms for single-source shortest paths (SSSP) type problems are not known due to the transitive-closure bottleneck (70), work-efficient algorithms that run in depth proportional to the diameter of the graph are known for the special cases considered in our benchmark. Several work-efficient parallel breadth-first search algorithms are known (11, 76, 21). On weighted graphs with integral edge weights, SSSP can be solved in work and depth (42). Parallel algorithms also exist for weighted graphs with positive edge weights (88, 87). SSSP on graphs with negative integer edge weights can be solved using Bellman-Ford (38), where the number of iterations depends on the diameter of the graph. Betweenness centrality from a single source can be computed using two breadth-first searches (30, 114).
In this paper, we present implementations of five SSSP problems that are based on graph search. We also include an algorithm to construct an -spanner which is based on computing low-diameter decompositions.
BFS, wBFS, Bellman-Ford, and Betweenness Centrality. Our implementations of BFS, Bellman-Ford, and betweenness centrality are based on the implementations in Ligra (114). Our wBFS implementation is based on our earlier work on Julienne (42). Algorithm 1 shows pseudocode for a frontier-based BFS implementation which synchronizes after each round of the BFS. The algorithm runs in and depth on the -, as vertices use test-and-set to non-deterministically acquire unvisited neighbors on the next frontier. Algorithm 4 shows pseudocode for a frontier-based version of Bellman-Ford which uses a priority-write to write the minimum distance to a vertex on each round and runs in work and depth on the -. Pseudocode for our betweenness centrality implementation is shown in Algorithm 2. It uses fetch-and-add to compute the total number of shortest-paths through vertices, and runs in work and depth on the -. Algorithm 3 shows pseudocode for our weighted BFS implementation from Julienne (42). The algorithm runs in work in expectation and depth w.h.p on the -, as vertices use priority-write to write the minimum distance to a neighboring vertex on each round. The algorithm uses a bucketing structure and generalized version of edgeMap called edgeMapData, which are discussed in the Julienne paper. The main change we made to these algorithms was to improve the cache-efficiency of the edgeMap implementation using the block-based version of edgeMap, described in Section 6.
Widest Path (Bottleneck Path). The Widest Path, or Bottleneck Path problem is to compute the maximum over all paths of the minimum weight edge on the path between a source vertex, , and . The algorithm is an important primitive, used for example in the Ford-Fulkerson flow algorithm (49, 38), as well as other flow algorithms (14). Sequentially, the algorithm can be solved as quickly as SSSP using a modified version of Dijkstra’s algorithm. We note that faster algorithms are known sequentially for sparse graphs (44). For positive integer-weighted graphs, the problem can also be solved using the the work-efficient bucketing data structure from Julienne (42). The buckets represent the width classes. The buckets are initialized with the out-neighbors of the source, , and the buckets are traversed using the decreasing order. This order specifies that the buckets are traversed from the largest bucket to the smallest bucket. Unlike the other applications in Julienne, using widest path is interesting since the bucket containing a vertex can only increase. The problem can also be solved using the Bellman-Ford approach, by performing computations over the semi-ring instead of the semi-ring. Other than these changes, the pseudocode for the problem is identical to that of Algorithms 3 and 4.
-Spanner. Computing graph spanners is a fundamental problem in combintaorial graph algorithms and graph theory (99). A graph is a -spanner of a graph if connected by a path, . Sequentially, classic results give elegant constructions of -spanners using edges, which are essentially the best possible assuming the girth conjecture (126). In this paper, we implement the recent spanner algorithm proposed by Miller, Peng, Xu, and Vladu (MPXV) (89). The construction results in an -spanner with expected size , and runs in work and depth on the -.
The MPXV spanner algorithm (Algorithm 5) uses the low-diameter decomposition (LDD) algorithm, which will be described in Section 5.2. It takes as input a parameter which controls the multiplicative stretch of the spanner. The idea is to first compute a LDD with (Line 3). The stretch of each ball is w.h.p., and so the algorithm includes all spanning tree edges generated by the LDD in the spanner (Line 4). Next, the algorithm handles inter-cluster edges by taking a single inter-cluster edge between a boundary vertex and its neighbors (Line 5). We note that this procedure is slightly different than the procedure in the MPXV paper, which adds a single edge between every boundary vertex of a cluster and each adjacent cluster. Our algorithm only adds a single edge between two clusters, while the MPXV algorithm may add multiple parallel edges between two clusters. Their argument bounding the stretch to for an edge spanning two clusters is still valid, since the endpoints can be first routed to the cluster centers, and then to the single edge that was selected between the two clusters.
5.2. Connectivity Problems
Low-Diameter Decomposition. Low-diameter decompositions (LDD) were first introduced in the context of distributed computing (7), and were later used in metric embedding, linear-system solvers, and parallel algorithms. Awerbuch presents a simple sequential algorithm based on ball growing that computes an decomposition (7). Miller, Peng, and Xu (MPX) present a work-efficient parallel algorithm that computes a decomposition. For each , the algorithm draws a start time, , from an exponential distribution with parameter . The clustering is done by assigning each vertex to the center which minimizes . This algorithm can be implemented by running a set of parallel breadth-first searches where the initial breadth-first search starts at the vertex with the largest start time, , and starting breadth-first searches from other once steps have elapsed. In this paper, we present an implementation of the MPX algorithm which computes an decomposition in expected work and depth w.h.p. on the -. Our implementation is based on the non-deterministic LDD implementation from Shun et al. (117). The main changes in our implementation are separating the LDD code from the connectivity implementation.
Algorithm 6 shows pseudocode for the modified version of the Miller-Peng-Xu algorithm from (117), which computes a decomposition in expected work and depth w.h.p. on the -. The algorithm allows ties to be broken arbitrarily when two searches visit a vertex in the time-step, and one can show that this only affects the number of cut edges by a constant factor (117). The algorithm first draws independent samples from (Line 1). Next, it computes the start times which is the difference between the maximum shifted value and (Line 2). Initially, all vertices are unvisited (Line 3). The algorithm performs ball-growing while all of the vertices are not yet visited (Line 10). On Line 11, it updates the current frontier with any vertices that are ready to start and have not yet been visited, and update the number of visited vertices with the size of the current frontier (Line 12). Finally, on Line 13, it uses edgeMap to traverse the out edges of the current frontier and non-deterministically acquire unvisited neighboring vertices.
Connectivity. Connectivity can be solved sequentially in linear work using breadth-first or depth-first search. Parallel algorithms for connectivity have a long history; we refer readers to (117) for a review of the literature. Early work on parallel connectivity discovered many natural algorithms which perform work (113, 8, 106, 101). A number of optimal parallel connectivity algorithms were discovered in subsequent years (50, 34, 57, 58, 102, 100, 117), but to the best of our knowledge the recent algorithm by Shun et al. is the only linear-work polylogarithmic-depth parallel algorithm that has been studied experimentally (117).
In this paper we implement the connectivity algorithm from Shun et al. (117), which runs in expected work and depth w.h.p. on the -. The implementation uses the work-efficient algorithm for low-diameter decomposition (LDD) (90) described above. One change we made to the implementation from (117) was to separate the LDD and contraction steps from the connectivity algorithm. Refactoring these sub-routines allowed us to express the main connectivity algorithm in Ligra in about 50 lines of code.
The connectivity algorithm from Shun et al. (117) (Algorithm 7) takes as input an undirected graph and a parameter . It first runs the LDD algorithm (Line 2) which decomposes the graph into clusters each with diameter , and inter-cluster edges in expectation. Next, it builds by contracting each cluster to a single vertex and adding inter-cluster edges while removing duplicate edges and isolated vertices (Line 3). It then checks if the contracted graph consists of isolated vertices (Line 4); if so, the clusters are the components, and it returns the mapping from vertices to clusters (Line 5). Otherwise, it recurses on the contracted graph (Line 7) and returns the connectivity labeling produced by assigning each vertex to the label assigned to its cluster in the recursive call (Line 9).
Spanning Forest. Finding spanning forests in parallel has been studied largely in conjunction with connectivity algorithms, since most parallel connectivity algorithms can naturally be modified to output a spanning forest (see (117) for a review of the literature). Our spanning forest algorithm (Algorithm 8) is based on the connectivity algorithm from Shun et al. (117) which we described earlier. The main difference in the spanning forest algorithm is to include all LDD edges at each level of the recuursion (Line 4). However, observe that the LDD edges after the topmost level of recursion are taken from a contracted graph, and need to be mapped back to some edge in the original graph realizing the contracted edge. We decide which edges in to add by maintaining a mapping from the edges in the current graph at some level of recursion to the original edge set. Initially this mapping, , is an identity map (Line 12). To compute the mapping to pass to the recursive call, we select any edge in the input graph that resulted in and map to (Line 8). In our implementation, we use a parallel hash table to select a single original edge per contracted edge.
Biconnectivity. Sequentially, biconnectivity can be solved using the Hopcroft-Tarjan algorithm (62). The algorithm uses depth-first search (DFS) to identify articulation points and requires work to label all edges with their biconnectivity label. It is possible to parallelize the sequential algorithm using a parallel DFS, however, the fastest parallel DFS algorithm is not work-efficient (2). Tarjan and Vishkin present the first work-efficient algorithm for biconnectivity (125) (as stated in the paper the algorithm is not work-efficient, but it can be made so by using a work-efficient connectivity algorithm). Another approach relies on the fact that biconnected graphs admit open ear decompositions to solve biconnectivity efficiently (82, 105).
In this paper, we implement the Tarjan-Vishkin algorithm for biconnectivity in expected work and depth on the -. Our implementation first computes connectivity labels using our connectivity algorithm, which runs in expected work and depth w.h.p. and picks an arbitrary source vertex from each component. Next, we compute a spanning forest rooted at these sources using breadth-first search, which runs in work and depth. We note that the connectivity algorithm can be modified to compute a spanning forest in the same work and depth as connectivity, which would avoid the breadth-first-search. We compute , , and for each vertex by running leaffix and rootfix sums on the spanning forests produced by BFS with fetch-and-add, which requires work and depth. Finally, we compute an implicit representation of the biconnectivity labels for each edge, using an idea from (16). This step computes per-vertex labels by removing all critical edges and computing connectivity on the remaining graph. The resulting vertex labels can be used to assign biconnectivity labels to edges by giving tree edges the connectivity label of the vertex further from the root in the tree, and assigning non-tree edges the label of either endpoint. Summing the cost of each step, the total work of this algorithm is in expectation and the total depth is w.h.p.
Algorithm 9 shows the Tarjan-Vishkin biconnectivity algorithm. We first compute a spanning forest of using the work-efficient connectivity algorithm, where the trees in the forest can be rooted arbitrarily (Line 2). Next, we compute a preorder numbering, , with respect to the roots (Line 3). We then compute for each and , which are the minimum and maximum preorder numbers respectively of all non-tree edges where is a vertex in ’s subtree. We also compute , the size of each vertex’s subtree. Note that we can determine whether the parent of a vertex is an articulation point by checking and . As in (16), we refer to this set of tree edges , where is an articulation point, as critical edges. The last step of the algorithm is to solve connectivity on the graph with all critical edges removed. Now, the biconnectivity label of an edge is the connectivity label of the vertex that is further from the root of the tree. The query data structure can thus report biconnectivity labels of edges in time using space, which is important for our implementations as storing a biconnectivity label per-edge explicitly would require a prohibitive amount of memory for large graphs. As the most costly step in this algorithm is to run connectivity, the algorithm runs in work in expectation and depth w.h.p. Our implementation described of the Tarjan-Vishkin algorithm runs in the same work but depth w.h.p. as it computes a spanning tree using BFS and performs leaffix and rootfix computations on this tree.
Minimum Spanning Forest. Borůvka gave the first known sequential and parallel algorithm for computing a minimum spanning forest (MSF) (29). Significant effort has gone into finding linear-work MSF algorithms both in the sequential and parallel settings (69, 34, 100). Unfortunately, the linear-work parallel algorithms are highly involved and do not seem to be practical. Significant effort has also gone into designing practical parallel algorithms for MSF; we discuss relevant experimental work in Section 7. Due to the simplicity of Borůvka, many parallel implementations of MSF use variants of it.
In this paper, we present an implementation of Borůvka’s algorithm that runs in work and depth on the -. Our implementation is based on a recent implementation of Borůvka by Zhou (130) that runs on the edgelist format. We made several changes to the algorithm which improve performance and allow us to solve MSF on graphs stored in the CSR/CSC format, as storing an integer-weighted graph in edgelist format would require well over 1TB of memory to represent the edges in the Hyperlink2012 graph alone. Our code uses an implementation of Borůvka that works over an edgelist; to make it efficient we ensure that the size of the lists passed to it are much smaller than . Our approach is to perform a constant number of filtering steps. Each filtering step solves an approximate ’th smallest problem in order to extract the lightest edges in the graph (or all remaining edges) and runs Borůvka on this subset of edges. We then filter the remaining graph, packing out any edges that are now in the same component. This idea is similar to the theoretically-efficient algorithm of Cole et al. (34), except that instead of randomly sampling edges, we select a linear number of the lowest weight edges. Each filtering step costs work and depth, but as we only perform a constant number of steps, they do not affect the work and depth asymptotically. In practice, most of the edges are removed after 3–4 filtering steps, and so the remaining edges can be copied into an edgelist and solved in a single Borůvka step. We also note that as the edges are initially represented in both directions, we can pack out the edges so that each undirected edge is only inspected once (we noticed that earlier edgelist-based implementations stored undirected edges in both directions).
Algorithm 10 shows an efficient implementation of Borůvka’s algorithm based on shortcutting using pointer-jumping instead of contraction (130). The Borůvka step is described in terms of the edgelist representation of a graph. We note that in practice, it is often the case that the MSF is contained in the first or so lowest weight edges. This motivates running several Borůvka steps on prefixes of the edges sorted by weight, an optimization which is also implemented by other experimental MSF algorithms (96, 116, 130). Each run computes a new set of edges that are now in the MSF which can be used to filter the remaining edges that are now shortcut by the newly added edges. After running a constant number of filter steps, we can run Borůvka on the remaining edges in the graph.
We now give a high-level description of the edgelist based Borůvka implementation and refer to (130) for a detailed explanation of the code. Each vertex is initially its own parent, all vertices are active and the initial forest is empty (Line 2). The algorithm runs over a series of rounds. In each round, we initialize cells for currently active vertices (Line 4). Next, we loop in parallel over all edges and perform a priority-write with min based on the weight on both endpoints of the edge (Lines 6 and 7). This writes the weight and index-id of a minimum-weight edge incident to a vertex into . On Line 8, we mark the roots of the forest (we break symmetry by marking the higher endpoint of an edge as the root of that edge). On Line 9 we filter vertices that did not have any vertices join their component and update the number of vertices that are still active. Next, we update the forest, adding all edges which won on either of their endpoints in (Line 10). We then use pointer-jumping to map every vertex to the id of its new component root (Line 11) and relabel all edges based on the new ids of each endpoint (Line 12). Finally, we filter any self-edges from the graph. We note that our implementation uses indirection over both the vertices and edges by maintaining a set of active vertices, of size and a set of active edge-ids. This helps improve performance in practice as we can allocate to have size proportional to the number of active vertices in each round, and as we can filter just the ids of the edges, instead of triples containing the two endpoints and the weight of each edge.
Strongly Connected Components. Tarjan’s algorithm is the textbook sequential algorithm for computing the strongly connected components (SCCs) of a directed graph (38). As it uses depth-first search, we currently do not know how to efficiently parallelize it (2). The current theoretical state-of-the-art for parallel SCC algorithms with polylogarithmic depth reduces the problem to computing the transitive closure of the graph. This requires work using combinatorial algorithms (51), which is significantly higher than the work done by sequential algorithms. As the transitive-closure based approach performs a significant amount of work even for moderately sized graphs, subsequent research on parallel SCC algorithms has focused on improving the work while potentially sacrificing depth (48, 37, 110, 23). Conceptually, these algorithms first pick a random pivot and use a reachability-oracle to identify the SCC containing the pivot. They then remove this SCC, which partitions the remaining graph into several disjoint pieces, and recurse on the pieces.
In this paper, we present the first implementation of the SCC algorithm from Blelloch et al. (23), shown in Algorithm 11. We refer the reader to Section 6.2 of (23) for proofs of correctness and its work and depth bounds. The algorithm is similar in spirit to randomized quicksort. On Line 2 we randomly permute the vertices and assign them to batches. On Line 3 we set the initial label for all vertices as and mark all vertices as not done. Now, we process the batches one at a time. For each batch, we compute , which are the vertices in the batch that are not yet done (Line 5). The next step calls MarkReachable from the centers on both and the transposed graph, (Lines 6–7). MarkReachable takes the set of centers and uses a breadth-first search to compute the sets (), which for a center includes all pairs for vertices that can reach through its in-edges (out-edges) that share the same label as . On Line 8, we mark all vertices that had a center visit them through both the in-search and out-search as done—these vertices are captured by some SCC. On Line 9, we deterministically set the label for vertices that were finished in this round. Finally, on Line 10, we set the labels for vertices that were not finished in this round but were visited by either the in or out-search.
Our implementation runs in in expected work and depth w.h.p. on the -. One of the challenges in implementing this SCC algorithm is how to compute reachability information from multiple vertices simultaneously and how to combine the information to (1) identify SCCs and (2) refine the subproblems of visited vertices. In our implementation, we explicitly store and , the forward and backward reachability sets for the set of centers that are active in the current phase, . The sets are represented as hash tables that store tuples of vertices and center IDs, , representing a vertex in the same subproblem as that is visited by a directed path from . We explain how to make the hash table technique practical in Section 6. The reachability sets are computed by running simultaneous breadth-first searches from all active centers. In each round of the BFS, we apply edgeMap to traverse all out-edges (or in-edges) of the current frontier. When we visit an edge we try to add ’s center IDs to . If succeeds in adding any IDs, it test-and-set’s a visited flag for , and returns it in the next frontier if the test-and-set succeeded. Each BFS requires at most rounds as each search adds the same labels on each round as it would have had it run in isolation.
After computing and , we deterministically assign (with respect to the random permutation of vertices generated at the start of the algorithm) vertices that we visited in this phase a new label, which is either the label of a refined subproblem or a unique label for the SCC they are contained in. We first intersect the two tables and perform, for any tuple contained in the intersection, a priority-write with min on the memory location corresponding to ’s SCC label with as the label. Next, for all pairs in we do a priority-write with min on ’s subproblem label, which ensures that the highest priority search that visited sets its new subproblem.
We implemented an optimized search for the first phase, which just runs two regular BFSs over the in-edges and out-edges from a single pivot and stores the reachability information in bit-vectors instead of hash-tables. It is well known that many directed real-world graphs have a single massive strongly connected component, and so with reasonable probability the first vertex in the permutation will find this giant component (32). We also implemented a ‘trimming’ optimization that is reported in the literature (85, 121), which eliminates trivial SCCs by removing any vertices that have zero in- or out-degree. We implement a procedure that recursively trims until no zero in- or out-degree vertices remain, or until a maximum number of rounds are reached.
5.3. Covering Problems
Maximal Independent Set. Maximal independent set (MIS) and maximal matching (MM) are easily solved in linear work sequentially using greedy algorithms. Many efficient parallel maximal independent set and matching algorithms have been developed over the years (71, 78, 4, 64, 22, 18). Blelloch et al. show that when the vertices (or edges) are processed in a random order, the sequential greedy algorithms for MIS and MM can be parallelized efficiently and give practical algorithms (22). Recently, Fischer and Noever showed an improved depth bound for this algorithm (47).
In this paper, we implement the rootset-based algorithm for MIS from Blelloch et al. (22) which runs in expected work and depth w.h.p. on the -. To the best of our knowledge this is the first implementation of the rootset-based algorithm; the implementations from (22) are based on processing appropriately-sized prefixes of an order generated by a random permutation . Our implementation of the rootset-based algorithm works on a priority-DAG defined by directing edges in the graph from the higher-priority endpoint to the lower-priority endpoint. On each round, we add all roots of the DAG into the MIS, compute , the neighbors of the rootset that are still active, and finally decrement the priorities of . As the vertices in are at arbitrary depths in the priority-DAG, we only decrement the priority along an edge , if . The algorithm runs in work as we process each edge once; the depth bound is as the priority-DAG has depth w.h.p. (47), and each round takes depth. We were surprised that this implementation usually outperforms the prefix-based implementation from (22), while also being simple to implement.
The rootset-based MIS algorithm from Blelloch et al. (22) is shown in Algorithm 12. In MIS, we first randomly order the vertices with a random permutation and compute the array , an array which for each vertex contains the number of neighbors that have higher priority than by calling according to (Lines 1–2). On Line 5 we compute the initial rootset, , which is the set of all vertices that have priority 0. In each round, the algorithm adds the roots to the independent set (Line 8), computes the set of removed vertices, which are neighbors of the rootset that are still active (). On Lines 10-11 we set the priority of the removed vertices to 0 and update the number of finished vertices. Finally, we compute the new rootset by decrementing the priority of all edges where and using a fetch-and-add, and returning true for a neighbor if we decrement its priority to 0.
Maximal Matching. Our maximal matching implementation is based on the prefix-based algorithm from (22) that takes expected work and depth w.h.p. on the - (using the improved depth shown in (47)). We had to make several modifications to run the algorithm on the large graphs in our experiments. The original code from (22) uses an edgelist representation, but we cannot directly use this implementation as uncompressing all edges would require a prohibitive amount of memory for large graphs. Instead, as in our MSF implementation, we simulate the prefix-based approach by performing a constant number of filtering steps. Each filter step packs out of the highest priority edges, randomly permutes them, and then runs the edgelist based algorithm on the prefix. After computing the new set of edges that are added to the matching, we filter the remaining graph and remove all edges that are incident to matched vertices. In practice, just 3–4 filtering steps are sufficient to remove essentially all edges in the graph. The last step uncompresses any remaining edges into an edgelist and runs the prefix-based algorithm. The filtering steps can be done within the work and depth bounds of the original algorithm.
The prefix-based maximal matching algorithm from Blelloch et al. (22) is shown in Algorithm 13. The algorithm first sets all vertices as unmatched, and initializes the matching to empty (Line 9). The algorithm runs over a set of rounds; each round selects a -prefix of the edges ( is the maximum number of neighboring edges an edge has), and runs the parallel greedy maximal matching algorithm (Line 1) on it. The parallel greedy algorithm repeatedly finds the set of edges that have the highest priority amongst all other edges incident to either endpoint (Line 4), adds them to the matching (Line 5), and filters the prefix based on the newly matched edges (Line 6). The edges matched by the greedy algorithm are returned to the MaximalMatching procedure (Line 12). We then recompute by removing the matched edges, and the one-hop neighborhood of the matched edges. Implementing this in practice can be done by lazily deleting edges, and using doubling to select the next prefix. We refer to (22) for the proof of the work and depth of this algorithm. Our actual implementation does several filtering steps before executing Algorithm 13 as described in Section 5. This does not affect the work and depth bounds.
Graph Coloring. As graph coloring is -hard to solve optimally, algorithms like greedy coloring, which guarantees a -coloring, are used instead in practice, and often use much fewer than colors on real-world graphs (127, 59). Jones and Plassmann (JP) parallelize the greedy algorithm using linear work, but unfortunately adversarial inputs exist for the heuristics they consider that may force the algorithm to run in depth. Hasenplaugh et al. introduce several heuristics that produce high-quality colorings in practice and also achieve provably low-depth regardless of the input graph. These include LLF (largest-log-degree-first), which processes vertices ordered by the log of their degree and SLL (smallest-log-degree-last), which processes vertices by removing all lowest log-degree vertices from the graph, coloring the remaining graph, and finally coloring the removed vertices. For LLF, they show that it runs in work and depth, where in expectation.
In this paper, we implement a synchronous version of Jones-Plassmann using the LLF heuristic in Ligra, which runs in work and depth on the -. The algorithm is implemented similarly to our rootset-based algorithm for MIS. In each round, after coloring the roots we use a fetch-and-add to decrement a count on our neighbors, and add the neighbor as a root on the next round if the count is decremented to 0.
Algorithm 14 shows a synchronous implementation of the parallel LLF-Coloring algorithm from (59). On Lines 1 and 2 we compute a random permutation , and compute which for each vertex contains the number of neighbors that have higher priority than based on comparing and breaking ties using . Initially, is the set of all vertices that have priority 0 (Line 4). In each round, we assign the lowest available color the roots and update (Line 6). We compute the next rootset by decrementing the priority of all edges where using a fetch-and-add, and returning any neighbor whose priority is decremented to 0 (Line 8).
Approximate Set Cover. The set cover problem can be modeled by a bipartite graph where sets and elements are vertices, with an edge between a set an element if and only if the set covers that element. Like graph coloring, the set cover problem is -hard to solve optimally, and a sequential greedy algorithm computes an -approximation in time for unweighted sets, and time for weighted sets, where and is the sum of the sizes of the sets (or the number of edges in the graph). There has been significant work on finding work-efficient parallel algorithms that achieves an -approximation (17, 103, 25, 26, 73).
Algorithm 15 shows pseudocode for the Blelloch et al. algorithm (25) which runs in work and depth on the -. We refer to (42) for a detailed explanation of the code, and give a high-level description of the algorithm here. The algorithm first buckets the sets based on their degree, placing a set covering elements into ’th bucket (Line 2). We process the buckets in decreasing order. In each round, we extract the highest bucket (Line 5) and pack out their adjacency lists to remove neighbors that may have been covered by prior rounds (Line 6). We then split the sets into , sets that continue in the round, and , sets that should be rebucketed based on whether the set still contains enough uncovered neighbors (Line 8). Next, we implement one step of MaNIS (25), which subselects a set of sets from that have little overlap. We select a random priority for each set in and try to acquire all neighbors of the set (Line 9) by doing a priority-write with min using the randomly chosen priority. On Line 10, we compute , the sets that won on more than of their neighbors, and add them to the cover (Line 11). Next, we reset memory locations for elements that were written-to by a set, but not acquired (Line 12). Finally we reinsert sets that did not win on enough neighbors (, or lost enough of their size () back into the bucket structure.
Our implementation of approximate set cover in this paper is based on the implementation from Julienne (42). The main change we made in this paper is to ensure that we correctly set random priorities on each round of the algorithm. Both the implementation in Julienne as well as an earlier implementation of the algorithm (26) use vertex IDs instead of picking random priorities for all sets that are active on a given round. This can cause very few vertices to be added on each round on meshes and other graphs with a large amount of symmetry. Instead, in our implementation, for , the active sets on a round, we generate a random permutation of and write these values into a pre-allocated dense array with size proportional to the number of sets. We give experimental details regarding this change in Section 7.
5.4. Substructure Problems
-core. -cores were defined independently by Seidman (111), and by Matula and Beck (83) who also gave a linear-time algorithm for computing the coreness value of all vertices, i.e. the maximum -core a vertex participates in. Anderson and Mayr showed that -core (and therefore coreness) is in NC for , but is P-complete for (6). The Matula and Beck algorithm is simple and practical—it first bucket-sorts vertices by their degree, and then repeatedly deletes the minimum-degree vertex. The affected neighbors are moved to a new bucket corresponding to their induced degree. As each edge in each direction and vertex is processed exactly once, the algorithm runs in work. In (42), the authors give a parallel algorithm based on bucketing that runs in expected work, and depth w.h.p. is the peeling-complexity of the graph, defined as the number of rounds to peel the graph to an empty graph where each peeling step removes all minimum degree vertices.
Our implementation of -core in this paper is based on the implementation from Julienne (42). One of the challenges to implementing the peeling algorithm for -core is efficiently computing the number of edges removed from each vertex that remains in the graph. A simple approach is to just fetch-and-add a counter per vertex, and update the bucket of the vertex based on this counter, however this incurs significant contention on real-world graphs with vertices with large degree. In order to make this step faster in practice, we implemented a work-efficient histogram which computes the number of edges removed from remaining vertices while incurring very little contention. We describe our histogram implementation in Section 6.
Algorithm 16 shows pseudocode for the work-efficient -core algorithm from Julienne (42) which computes the coreness values of all vertices. The algorithm initializes the initial coreness values to the degree of each vertex (Line 2), and inserts the vertices into a bucketing data-structure based on their degree (Line 3). In each round, while all of the vertices have not yet been processed the algorithm removes the vertices in the minimum bucket (Line 5), computes the number of edges removed from each neighbor using a histogram (Line 7) and finally updates the buckets of affected neighbors (Line 8). We return the array , which contains the coreness values of each vertex at the end of the algorithm.
Approximate Densest Subgraph. The densest subgraph problem is to find a subgraph of an undirected graph with the highest density (the density of a subgraph is the number of edges in the subgraph divided by the number of vertices). The problem is a classic graph optimization problem that admits exact polynomial-time solutions using either a reduction to flow (52) or LP-rounding (33). In his paper, Charikar also gives a simple work 2-approximation algorithm based on computing a degeneracy ordering of the graph, and taking the maximum density subgraph over all suffixes of the degeneracy order33 3 We note that the -approximation be work-efficiently solved in the same depth as our -core algorithm by augmenting the -core algorithm to return the order in which vertices are peeled. Computing the maximum density subgraph over suffixes of the degeneracy order can be done using scan.. The problem has also received attention in parallel models of computation (13, 12). Bahmani et al. give a -approximation running in rounds of MapReduce (13). Subsequently, Bahmani et al. (12) showed that a can be found in rounds of MapReduce by using the multiplicative-weights approach on the dual of the natural LP for densest subgraph. To the best our knowledge, it is currently open whether the densest subgraph problem can be exactly solved in .
In this paper, we implement the elegant -approximation algorithm of Bahmani et al. (Algorithm 17). Our implementation of the algorithm runs in work and depth. The algorithm starts with a candidate densest subgraph, , consisting of all vertices, and an empty densest subgraph (Line 7). It also maintains an array with the induced degree of each vertex in , which is initially just its degree in (Line 1). The main loop iteratively peels vertices with degree below the density threshold in the current candidate subgraph (Lines 8–13). Specifically, it finds all vertices with induced degree less than (Line 9), calls edgeMap, which updates the induced degrees array (Line 10) and finally removes them from (Line 11). If the density of the updated subgraph is greater than the density of , the algorithm updates to be .
Bahmani et al. show that this algorithm removes a constant factor of the vertices in each round. However, they do not consider the work of the algorithm in the MapReduce model. We briefly sketch how the algorithm can be implemented in work and depth. Instead of computing the density of the current subgraph by scanning all edges, we maintain it explicitly in an array, (Line 1), and update it as vertices are removed from . Each round of the algorithm does work proportional to vertices in to compute (Line 9) but since decreases by a constant factor in each round the work of these steps is over all rounds. Computing the new density can be done by computing the number of edges between and , which only requires scanning edges incident to vertices in using edgeMap (Line 10). Therefore, the edges incident to a vertex are scanned exactly once, in the round when it is included in , and so the algorithm performs work. The depth is since there are rounds each of which perform a filter and edgeMap which both run in depth. In practice, we found that since a large number of vertices are removed in each round, using fetch-and-add can cause contention, especially on graphs containing vertices with high degrees. Instead, our implementation uses a work-efficient histogram procedure (see Section 6) which updates the degrees while incurring very little contention.
Triangle Counting. Triangle counting has received significant recent attention due to its numerous applications in Web and social network analysis. There have been dozens of papers on sequential triangle counting (65, 5, 109, 108, 75, 95, 97). The fastest algorithms rely on matrix multiplication and run in either or work, where is the best matrix multiplication exponent (65, 5). The fastest algorithm that does not rely matrix multiplication requires work (109, 108, 75), which also turns out to be much more practical. Parallel algorithms with work have been designed (119, 1, 77), with Shun and Tangwongsan (119) showing an algorithm that requires depth on the .44 4 The algorithm in (119) was described in the Parallel Cache Oblivious model, with a depth of .
The implementation from (119) parallelizes Latapy’s compact-forward algorithm, which creates a directed graph where an edge is kept in iff . Although triangle counting can be done directly on the undirected graph in the same work and depth asymptotically, directing the edges helps reduce work, and ensures that every triangle is counted exactly once.
In this paper we implement the triangle counting algorithm described in (119). We had to make several significant changes to the implementation in order to run efficiently on large compressed graphs. First, we parallelized the creation of the directed graph; this step creates a directed graph encoded in the parallel-byte format in work and depth. We also parallelized the merge-based intersection algorithm to make it work in the parallel-byte format. We give more details on these techniques in Section 6.
5.5. Eigenvector Problems
PageRank. PageRank is a centrality algorithm first used at Google to rank webpages (31). The algorithm takes a graph , a damping factor and a constant which controls convergence. Initially, the PageRank of each vertex is . On each iteration, the algorithm updates the PageRanks of the vertices using the following equation:
This update step can be implemented using a single sparse-matrix vector multiplication call (implementable using, for example, edgeMap). The algorithm implemented in this paper follows the implementation of PageRank described in Ligra (114). We note that many PageRank implementations in the wild actually implement an algorithm called PageRank-Delta, which modified PageRank by only activating a vertex if its PageRank value has changed sufficiently. However, we are not aware of any bounds on the work and depth of this algorithm, and therefore chose to implement the classic PageRank.
The main modification we made to the implementation from Ligra was to implement the dense iterations of the algorithm using a reduction primitive, which can be carried out over the incoming neighbors of a vertex in parallel, without needing a fetch-and-add instruction. Each iteration of our implementation requires work and depth.
6. Implementations and Techniques
In this section, we introduce several general implementation techniques and optimizations that we use in our algorithms. The techniques include a fast histogram implementation useful for reducing contention in the -core algorithm, a cache-friendly sparse edgeMap implementation that we call edgeMapBlocked, and compression techniques used to efficiently parallelize algorithms on massive graphs.
6.1. A Work-efficient Histogram Implementation
Our initial implementation of the peeling-based algorithm for -core algorithm suffered from poor performance due to a large amount of contention incurred by fetch-and-adds on high-degree vertices. This occurs as many social-networks and web-graphs have large maximum degree, but relatively small degeneracy, or largest non-empty core (labeled in Table 2). For these graphs, we observed that many early rounds, which process vertices with low coreness perform a large number of fetch-and-adds on memory locations corresponding to high-degree vertices, resulting in high contention (115). To reduce contention, we designed a work-efficient histogram implementation that can perform this step while only incurring contention w.h.p. The Histogram primitive takes a sequence of pairs, and an associative and commutative operator and computes a sequence of pairs, where each key only appears once, and its associated value is the sum of all values associated with keys in the input, combined with respect to .
A useful example of histogram to consider is summing for each for a vertexSubset , the number of edges where (i.e., the number of incoming neighbors from the frontier). This operation can be implemented by running histogram on a sequence where each appears once per edge as a tuple using the operator . One theoretically efficient implementation of histogram is to simply semisort the pairs using the work-efficient semisort algorithm from (56). The semisort places pairs from the sequence into a set of heavy and light buckets, where heavy buckets contain a single key that appears many times in the input sequence, and light buckets contain at most distinct keys keys, each of which appear at most times w.h.p. (heavy and light keys are determined by sampling). We compute the reduced value for heavy buckets using a standard parallel reduction. For each light bucket, we allocate a hash table, and hash the keys in the bucket in parallel to the table, combining multiple values for the same key using . As each key appears at most times w.h.p, we incur at most ) contention w.h.p. The output sequence can be computed by compacting the light tables and heavy arrays.
While the semisort implementation is theoretically efficient, it requires a likely cache miss for each key when inserting into the appropriate hash table. To improve cache performance in this step, we implemented a work-efficient algorithm with depth based on radix sort. Our implementation is based on the parallel radix sort from PBBS (116). As in the semisort, we first sample keys from the sequence and determine the set of heavy-keys. Instead of directly moving the elements into light and heavy buckets, we break up the input sequence into blocks, each of size , and sequentially sort the keys within a block into light and heavy buckets. Within the blocks, we reduce all heavy keys into a single value and compute an array of size which holds the starting offset of each bucket within the block. Next, we perform a segmented-scan (19) over the arrays of the blocks to compute the sizes of the light buckets, and the reduced values for the heavy-buckets, which only contain a single key. Finally, we allocate tables for the light buckets, hash the light keys in parallel over the blocks and compact the light tables and heavy keys into the output array. Each step runs in work and depth. Compared to the original semisort implementation, this version incurs fewer cache misses because the light keys per block are already sorted and consecutive keys likely go to the same hash table, which fits in cache. We compared our times in the histogram-based version of -core and the fetch-and-add-based version of -core and saw between a 1.1–3.1x improvement from using the histogram.
6.2. edgeMapBlocked
One of the core primitives used by our algorithms is edgeMap (described in Section 3). The push-based version of edgeMap, edgeMapSparse, takes a frontier and iterates over all edges incident to it. It applies an update function on each edge that returns a boolean indicating whether or not the neighbor should be included in the next frontier. The standard implementation of edgeMapSparse first computes prefix-sums of to compute offsets, allocates an array of size , and iterates over all in parallel, writing the ID of the neighbor to the array if the update function returns true, and otherwise. It then filters out the values in the array to produce the output vertexSubset.
In real-world graphs, , the number of unique neighbors incident to the current frontier is often much smaller than . However, edgeMapSparse will always perform writes and incur a proportional number of cache misses, despite the size of the output being at most . More precisely, the size of the output is at most , where is the number of live neighbors of , where a live neighbor is a neighbor of the current frontier for which returns true. To reduce the number of cache misses we incur in the push-based traversal, we implemented a new version of edgeMapSparse that performs at most writes that we call edgeMapBlocked. The idea behind edgeMapBlocked is to logically break the edges incident to the current frontier up into a set of blocks, and iterate over the blocks sequentially, packing live neighbors, compactly for each block. We then simply prefix-sum the number of live neighbors per-block, and compact the block outputs into the output array.
We now describe a theoretically efficient implementation of edgeMapBlocked (Algorithm 18). As in edgeMapSparse, we first compute an array of offsets (Line 1) by prefix summing the degrees of . We process the edges incident to this frontier in blocks of size . As we cannot afford to explicitly write out the edges incident to the current frontier to block them, we instead logically assign the edges to blocks. Each block searches for a range of vertices to process with edges; the ’th block binary searches the offsets array to find the vertex incident to the start of the ’th edge, storing the result into (Lines 4–5). The vertices that block must process are therefore between and . We note that multiple blocks can be assigned to process the edges incident to a high-degree vertex. Next, we allocate an intermediate array of size (Line 6), but do not initialize the memory, and an array that stores the number of live neighbors found by each block (Line 7). Next, we process the blocks in parallel by sequentially applying to each edge in the block and compactly writing any live neighbors to (Line 9), and write the number of live neighbors to (Line 10). Finally, we do a prefix sum on , which gives offsets into an array of size proportional to the number of live neighbors, and copy the live neighbors in parallel to , the output array (Line 11).
We found that this optimization helps the most in algorithms where there is a significant imbalance between the size of the output of each edgeMap, and . For example, in weighted BFS, relatively few of the edges actually relax a neighboring vertex, and so the size of the output, which contains vertices that should be moved to a new bucket, is usually much smaller than the total number of edges incident to the frontier. In this case, we observed as much as a 1.8x improvement in running time by switching from edgeMapSparse to edgeMapBlocked.
6.3. Techniques for overlapping searches
In this section, we describe how we compute and update the reachability labels for vertices that are visited in a phase of our SCC algorithm. Recall that each phase performs a graph traversal from the set of active centers on this round, , and computes for each center , all vertices in the weakly-connected component for the subproblem of that can be reached by a directed path from it. We store this reachability information as a set of pairs in a hash-table, which represent the fact that can be reached by a directed path from . A phase performs two graph traversals from the centers to compute and , the out-reachability set and in-reachability sets respectively. Each traversal allocates an initial hash table and runs rounds of edgeMap until no new label information is added to the table.
The main challenge in implementing one round in the traversal is (1) ensuring that the table has sufficient space to store all pairs that will be added this round, and (2) efficiently iterating over all of the pairs associated with a vertex. We implement (1) by performing a parallel reduce to sum over vertices , the current frontier, the number of neighbors in the same subproblem, multiplied by the number of distinct labels currently assigned to . This upper-bounds the number of distinct labels that could be added this round, and although we may overestimate the number of actual additions, we will never run out of space in the table. We update the number of elements currently in the table during concurrent insertions by storing a per-processor count which gets incremented whenever the processor performs a successful insertion. The counts are then summed together at the end of a round and used to update the count of the number of elements in the table.
One simple implementation of (2) is to simply allocate space for every vertex, as the maximum number of centers that visit any vertex during a phase is at most w.h.p. However, this will waste a significant amount of space, as most vertices are visited just a few times. Instead, our implementation stores pairs in the table for visited vertices , and computes hashes based only on the ID of . As each vertex is only expected to be visited a constant number of times during a phase, the expected probe length is still a constant. Storing the pairs for a vertex in the same probe-sequence is helpful for two reasons. First, we may incur fewer cache misses than if we had hashed the pairs based on both entries, as multiple pairs for a vertex can fit in the same cache line. Second, storing the pairs for a vertex along the same probe sequence makes it extremely easy to find all pairs associated with a vertex , as we simply perform linear-probing, reporting all pairs that have as their key until we hit an empty cell. Our experiments show that this technique is practical, and we believe that it may have applications in similar algorithms, such as computing least-element lists or FRT trees in parallel (23, 24).
6.4. Primitives on Compressed Graphs
Many of our algorithms are concisely expressed using fundamental primitives such as map, map-reduce, filter, pack, and intersection. To run our algorithms without any modifications on compressed graphs, we wrote new implementations of these primitives using using the parallel-byte format from Ligra+, some of which required some new techniques in order to be theoretically efficient. We first review the byte and parallel-byte formats from (118). In byte coding, we store a vertex’s neighbor list by difference encoding consecutive vertices, with the first vertex difference encoded with respect to the source. Decoding is done by sequentially uncompressing each difference, and summing the differences into a running sum which gives the ID of the next neighbor. As this process is sequential, graph algorithms using the byte format that map over the neighbors of a vertex will require depth. The parallel-byte format from Ligra+ breaks the neighbors of a high-degree vertex into blocks, where each block contains a constant number of neighbors. Each block is difference encoded with respect to the source. As each block can have a different size, it also stores offsets that point to the start of each block. The format stores the blocks in a neighbor list in sorted order.
We now describe efficient implementations of primitives used by our algorithms. All descriptions are given for neighbor lists coded in the parallel-byte format. The Map primitive takes as input neighbor list , and a map function , and applies to each ID in . This can be implemented with a parallel-for loop across the blocks, where each iteration decodes its block sequentially. Our implementation of map runs in work and depth. Map-Reduce takes as input a neighbor list , a map function and a binary associative function and returns the sum of the mapped elements with respect to . We perform map-reduce similarly by first mapping over the blocks, then sequentially reducing over the mapped values in each block. We store the accumulated value on the stack or in an allocated array if the number of blocks is large enough. Finally, we reduce the accumulated values using to compute the output. Our implementation of map-reduce runs in work and depth.
Filter takes as input a neighbor list , a predicate , and an array into which the vertices satisfying are written, in the same order as in . Our implementation of filter also takes as input an array , which is an array of size space for lists larger than a constant threshold, and null otherwise. In the case where is large, we implement the filter by first decoding into in parallel; each block in has an offset into as every block except possibly the last block contains the same number of vertex IDs. We then filter into the output array . In the case where is small we just run the filter sequentially. Our implementation of filter runs in work and depth. Pack takes as input a neighbor list and a predicate function, and packs , keeping only vertex IDs that satisfied . Our implementation of pack takes as input an array , which an array of size for lists larger than a constant threshold, and null otherwise. In the case where is large, we first decode in parallel into the first cells of . Next, we filter these vertices into the second cells of , and compute the new length of . Finally, we recompress the blocks in parallel by first computing the compressed size of each new block. We prefix-sum the sizes to calculate offsets into the array and finally compress the new blocks by writing each block starting at its offset. When is small we just pack sequentially. We make use of the pack and filter primitives in our implementations of maximal matching, minimum spanning forest, and triangle counting. Our implementation of pack runs in work and depth.
The Intersection primitive takes as input two neighbor lists and and computes the size of the intersection of and (). We implement an algorithm similar to the optimal parallel intersection algorithm for sorted lists. As the blocks are compressed, our implementation works on the first element of each block, which can be quickly decoded. We refer to these elements as block starts. If the number of blocks in both lists sum to less than a constant, we intersect them sequentially. Otherwise, we take the start of the middle block in , and binary search over the starts of to find the first block whose start is less than or equal to . Note that as the closest value less than or equal to could be in the middle of the block, the subproblems we generate must to consider elements in the two adjoining blocks of each list, which adds an extra constant factor of work in the base case. Our implementation of intersection runs in work and depth.
7. Experiments
In this section, we describe our experimental results on a set of real-world graphs and also discuss related experimental work. Tables 3 and 4 show the running times for our implementations on our graph inputs. For compressed graphs, we use the compression schemes from Ligra+ (118), which we extended to ensure theoretical efficiency. We describe these modifications and also other statistics about our algorithms (e.g., number of colors used, number of SCCs, etc.) in section A.
7.1. Experimental Setup and Graph Inputs
Experimental Setup. We run all of our experiments on a 72-core Dell PowerEdge R930 (with two-way hyper-threading) with Intel 18-core E7-8867 v4 Xeon processors (with a 4800MHz bus and 45MB L3 cache) and 1TB of main memory. Our programs use Cilk Plus to express parallelism and are compiled with the g++ compiler (version 5.4.1) with the -O3 flag. By using Cilk’s work-stealing scheduler we are able obtain an expected running time of for an algorithm with work and depth on processors (27). For the parallel experiments, we use the command numactl -i all to balance the memory allocations across the sockets. All of the speedup numbers we report are the running times of our parallel implementation on 72-cores with hyper-threading over the running time of the implementation on a single thread.
| Graph Dataset | Num. Vertices | Num. Edges | |||
|---|---|---|---|---|---|
| LiveJournal | 4,847,571 | 68,993,773 | 16 | ||
| LiveJournal-Sym | 4,847,571 | 85,702,474 | 20 | 3480 | 372 |
| com-Orkut | 3,072,627 | 234,370,166 | 9 | 5,667 | 253 |
| 41,652,231 | 1,468,365,182 | 65* | |||
| Twitter-Sym | 41,652,231 | 2,405,026,092 | 23* | 14,963 | 2488 |
| 3D-Torus | 1,000,000,000 | 6,000,000,000 | 1500* | 1 | 6 |
| ClueWeb | 978,408,098 | 42,574,107,469 | 821* | ||
| ClueWeb-Sym | 978,408,098 | 74,744,358,622 | 132* | 106,819 | 4244 |
| Hyperlink2014 | 1,724,573,718 | 64,422,807,961 | 793* | ||
| Hyperlink2014-Sym | 1,724,573,718 | 124,141,874,032 | 207* | 58,711 | 4160 |
| Hyperlink2012 | 3,563,602,789 | 128,736,914,167 | 5275* | ||
| Hyperlink2012-Sym | 3,563,602,789 | 225,840,663,232 | 331* | 130,728 | 10565 |
Graph Data. To show how our algorithms perform on graphs at different scales, we selected a representative set of real-world graphs of varying sizes. Most of the graphs are Web graphs and social networks—low diameter graphs that are frequently used in practice. To test our algorithms on large diameter graphs, we also ran our implementations 3-dimensional tori where each vertex is connected to its 2 neighbors in each dimension.
We list the graphs used in our experiments, along with their size, approximate diameter, peeling complexity (42), and degeneracy (for undirected graphs) in Table 2. LiveJournal is a directed graph of the social network obtained from a snapshot in 2008 (28). com-Orkut is an undirected graph of the Orkut social network. Twitter is a directed graph of the Twitter network, where edges represent the follower relationship (74). ClueWeb is a Web graph from the Lemur project at CMU (28). Hyperlink2012 and Hyperlink2014 are directed hyperlink graphs obtained from the WebDataCommons dataset where nodes represent web pages (86). 3D-Torus is a 3-dimensional torus with 1B vertices and 6B edges. We mark symmetric (undirected) versions of the directed graphs with the suffix -Sym. We create weighted graphs for evaluating weighted BFS, Borůvka, widest path, and Bellman-Ford by selecting edge weights between uniformly at random. We process LiveJournal, com-Orkut, Twitter, and 3D-Torus in the uncompressed format, and ClueWeb, Hyperlink2014, and Hyperlink2012 in the compressed format.
| Application | LiveJournal-Sym | com-Orkut | Twitter-Sym | 3D-Torus | ||||||||
| (1) | (72h) | (SU) | (1) | (72h) | (SU) | (1) | (72h) | (SU) | (1) | (72h) | (SU) | |
| Breadth-First Search (BFS) | 0.59 | 0.018 | 32.7 | 0.41 | 0.012 | 34.1 | 5.45 | 0.137 | 39.7 | 301 | 5.53 | 54.4 |
| Integral-Weight SSSP (weighted BFS) | 1.45 | 0.107 | 13.5 | 2.03 | 0.095 | 21.3 | 33.4 | 0.995 | 33.5 | 437 | 18.1 | 24.1 |
| General-Weight SSSP (Bellman-Ford) | 3.39 | 0.086 | 39.4 | 3.98 | 0.168 | 23.6 | 48.7 | 1.56 | 31.2 | 6280 | 133 | 47.2 |
| Single-Source Widest Path (Bellman-Ford) | 3.48 | 0.090 | 38.6 | 4.39 | 0.098 | 44.7 | 42.4 | 0.749 | 56.6 | 580 | 9.7 | 59.7 |
| Single-Source Betweenness Centrality (BC) | 1.66 | 0.049 | 33.8 | 2.52 | 0.057 | 44.2 | 26.3 | 0.937 | 28.0 | 496 | 12.5 | 39.6 |
| -Spanner | 1.31 | 0.041 | 31.9 | 2.34 | 0.046 | 50.8 | 41.5 | 0.768 | 54.0 | 380 | 11.7 | 32.4 |
| Low-Diameter Decomposition (LDD) | 0.54 | 0.027 | 20.0 | 0.33 | 0.019 | 17.3 | 8.48 | 0.186 | 45.5 | 275 | 7.55 | 36.4 |
| Connectivity | 1.01 | 0.029 | 34.8 | 1.36 | 0.031 | 43.8 | 34.6 | 0.585 | 59.1 | 300 | 8.71 | 34.4 |
| Spanning Forest | 1.11 | 0.035 | 31.7 | 1.84 | 0.047 | 39.1 | 43.2 | 0.818 | 52.8 | 334 | 10.1 | 33.0 |
| Biconnectivity | 5.36 | 0.261 | 20.5 | 7.31 | 0.292 | 25.0 | 146 | 4.86 | 30.0 | 1610 | 59.6 | 27.0 |
| Strongly Connected Components (SCC)* | 1.61 | 0.116 | 13.8 | 13.3 | 0.495 | 26.8 | ||||||
| Minimum Spanning Forest (MSF) | 3.64 | 0.204 | 17.8 | 4.58 | 0.227 | 20.1 | 61.8 | 3.02 | 20.4 | 617 | 23.6 | 26.1 |
| Maximal Independent Set (MIS) | 1.18 | 0.034 | 34.7 | 2.23 | 0.052 | 42.8 | 34.4 | 0.759 | 45.3 | 236 | 4.44 | 53.1 |
| Maximal Matching (MM) | 2.42 | 0.095 | 25.4 | 4.65 | 0.183 | 25.4 | 46.7 | 1.42 | 32.8 | 403 | 11.4 | 35.3 |
| Graph Coloring | 4.69 | 0.392 | 11.9 | 9.05 | 0.789 | 11.4 | 148 | 6.91 | 21.4 | 350 | 11.3 | 30.9 |
| Approximate Set Cover | 4.65 | 0.613 | 7.58 | 4.51 | 0.786 | 5.73 | 66.4 | 3.31 | 20.0 | 1429 | 40.2 | 35.5 |
| -core | 3.75 | 0.641 | 5.85 | 8.32 | 1.33 | 6.25 | 110 | 6.72 | 16.3 | 753 | 6.58 | 114.4 |
| Approximate Densest Subgraph | 2.89 | 0.052 | 55.5 | 4.71 | 0.081 | 58.1 | 76.0 | 1.14 | 66.6 | 95.4 | 1.59 | 60.0 |
| Triangle Counting (TC) | 13.5 | 0.342 | 39.4 | 78.1 | 1.19 | 65.6 | 1920 | 23.5 | 81.7 | 168 | 6.63 | 25.3 |
| PageRank Iteration | 0.861 | 0.012 | 71.7 | 1.28 | 0.018 | 71.1 | 24.16 | 0.453 | 53.3 | 107 | 2.25 | 47.5 |
| Application | ClueWeb-Sym | Hyperlink2014-Sym | Hyperlink2012-Sym | ||||||
| (1) | (72h) | (SU) | (1) | (72h) | (SU) | (1) | (72h) | (SU) | |
| Breadth-First Search (BFS) | 106 | 2.29 | 46.2 | 250 | 4.50 | 55.5 | 576 | 8.44 | 68.2 |
| Integral-Weight SSSP (weighted BFS) | 736 | 14.4 | 51.1 | 1390 | 22.3 | 62.3 | 3770 | 58.1 | 64.8 |
| General-Weight SSSP (Bellman-Ford) | 1050 | 16.2 | 64.8 | 1460 | 22.9 | 63.7 | 4010 | 59.4 | 67.5 |
| Single-Source Widest Path (Bellman-Ford) | 849 | 11.8 | 71.9 | 1211 | 16.8 | 72.0 | 3210 | 48.4 | 66.3 |
| Single-Source Betweenness Centrality (BC) | 569 | 27.7 | 20.5 | 866 | 16.3 | 53.1 | 2260 | 37.1 | 60.9 |
| -Spanner | 613 | 9.79 | 62.6 | 906 | 14.3 | 63.3 | 2390 | 36.3 | 65.8 |
| Low-Diameter Decomposition (LDD) | 176 | 3.62 | 48.6 | 322 | 6.84 | 47.0 | 980 | 16.6 | 59.0 |
| Connectivity | 381 | 6.01 | 63.3 | 710 | 11.2 | 63.3 | 1640 | 25.0 | 65.6 |
| Spanning Forest | 936 | 18.2 | 51.4 | 1319 | 22.4 | 58.8 | 2420 | 35.8 | 67.5 |
| Biconnectivity | 2250 | 48.7 | 46.2 | 3520 | 71.5 | 49.2 | 9860 | 165 | 59.7 |
| Strongly Connected Components (SCC)* | 1240 | 38.1 | 32.5 | 2140 | 51.5 | 41.5 | 8130 | 185 | 43.9 |
| Minimum Spanning Forest (MSF) | 2490 | 45.6 | 54.6 | 3580 | 71.9 | 49.7 | 9520 | 187 | 50.9 |
| Maximal Independent Set (MIS) | 551 | 8.44 | 65.2 | 1020 | 14.5 | 70.3 | 2190 | 32.2 | 68.0 |
| Maximal Matching (MM) | 1760 | 31.8 | 55.3 | 2980 | 48.1 | 61.9 | 7150 | 108 | 66.2 |
| Graph Coloring | 2050 | 49.8 | 41.1 | 3310 | 63.1 | 52.4 | 8920 | 158 | 56.4 |
| Approximate Set Cover | 1490 | 28.1 | 53.0 | 2040 | 37.6 | 54.2 | 5320 | 90.4 | 58.8 |
| -core | 2370 | 62.9 | 37.6 | 3480 | 83.2 | 41.8 | 8515 | 184 | 46.0 |
| Approximate Densest Subgraph | 1380 | 19.6 | 70.4 | 1721 | 24.3 | 70.8 | 4420 | 61.4 | 71.9 |
| Triangle Counting (TC) | 13997 | 204 | 68.6 | — | 480 | — | — | 1168 | — |
| PageRank Iteration | 256.1 | 3.49 | 73.3 | 385 | 5.17 | 74.4 | 973 | 13.1 | 74.2 |
7.2. SSSP Problems
Our BFS, weighted BFS, Bellman-Ford, and betweenness centrality implementations achieve between a 13–67x speedup across all inputs. We ran all of our shortest path experiments on the symmetrized versions of the graph. Our widest path implementation achieves between 38–72x speedup across all inputs, and our spanner implementation achieves between 31–65x speedup across all inputs. We ran our spanner code with . Our experiments show that our weighted BFS and Bellman-Ford implementations perform as well as or better than our prior implementations from Julienne (42). Our running times for BFS and betweenness centrality are the same as the times of the implementations in Ligra (114). We note that our running times for weighted BFS on the Hyperlink graphs are larger than the times reported in Julienne. This is because the shortest-path experiments in Julienne were run on directed version of the graph, where the average vertex can reach many fewer vertices than on the symmetrized version. We set a flag for our weighted BFS experiments on the ClueWeb and Hyperlink graphs that lets the algorithm switch to a dense edgeMap once the frontiers are sufficiently dense, which lets the algorithm run within half of the RAM on our machine. Before this change, our weighted BFS implementation would request a large amount of amount of memory when processing the largest frontiers which then caused the graph to become partly evicted from the page cache. For widest path, the times we report are for the Bellman-Ford version of the algorithm, which we were surprised to find is consistently 1.1–1.3x faster than our algorithm based on bucketing. We observe that our spanner algorithm is only slightly more costly than computing connectivity on the same input.
In an earlier paper (42), we compared the running time of our weighted BFS implementation to two existing parallel shortest path implementations from the GAP benchmark suite (15) and Galois (80), as well as a fast sequential shortest path algorithm from the DIMACS shortest path challenge, showing that our implementation is between 1.07–1.1x slower than the -stepping implementation from GAP, and 1.6–3.4x faster than the Galois implementation. Our old version of Bellman-Ford was between 1.2–3.9x slower than weighted BFS; we note that after changing it to use the edgeMapBlocked optimization, it is now competitive with weighted BFS and is between 1.2x faster and 1.7x slower on our graphs with the exception of 3D-Torus, where it performs 7.3x slower than weighted BFS, as it performs work on this graph.
7.3. Connectivity Problems
Our low-diameter decomposition (LDD) implementation achieves between 17–59x speedup across all inputs. We fixed to in all of the codes that use LDD. The running time of LDD is comparable to the cost of a BFS that visits most of the vertices. We are not aware of any prior experimental work that reports the running times for an LDD implementation.
Our work-efficient implementation of connectivity and spanning forest achieve 25–57x speedup and 31–67x speedup across all inputs, respectively. We note that our implementation does not assume that vertex IDs in the graph are randomly permuted and always generates a random permutation, even on the first round, as adding vertices based on their original IDs can result in poor performance (for example on 3D-Torus). There are several existing implementations of fast parallel connectivity algorithms (98, 116, 117, 121), however, only the implementation from (117), which presents the connectivity algorithm that we implement in this paper, is theoretically-efficient. The implementation from Shun et al. was compared to both the Multistep (121) and Patwary et al. (98) implementations, and shown to be competitive on a broad set of graphs. We compared our connectivity implementation to the work-efficient connectivity implementation from Shun et al. on our uncompressed graphs and observed that our code is between 1.2–2.1x faster in parallel. Our spanning forest implementation is slightly slower than connectivity due to having to maintain a mapping between the current edge set and the original edge set.
Despite our biconnectivity implementation having depth, our implementation achieves between a 20–59x speedup across all inputs, as the diameter of most of our graphs is extremely low. Our biconnectivity implementation is about 3–5 times slower than running connectivity on the graph, which seems reasonable as our current implementation performs two calls to connectivity, and one breadth-first search. There are a several existing implementations of biconnectivity. Cong and Bader (35) parallelize the Tarjan-Vishkin algorithm and demonstrated speedup over the Hopcroft-Tarjan (HT) algorithm. Edwards and Vishkin (45) also implement the Tarjan-Vishkin algorithm using the XMT platform, and show that their algorithm achieves good speedups. Slota and Madduri (120) present a BFS-based biconnectivity implementation which requires work in the worst-case, but behaves like a linear-work algorithm in practice. We ran the Slota and Madduri implementation on 36 hyper-threads allocated from the same socket, the configuration on which we observed the best performance for their code, and found that our implementation is between 1.4–2.1x faster than theirs. We used a DFS-ordered subgraph corresponding to the largest connected component to test their code, which produced the fastest times. Using the original order of the graph affects the running time of their implementation, causing it to run between 2–3x slower as the amount of work performed by their algorithm depends on the order in which vertices are visited.
Our strongly connected components implementation achieves between a 13–43x speedup across all inputs. Our implementation takes a parameter , which is the base of the exponential rate at which we grow the number of centers added. We set between – for our experiments and note that using a larger value of can improve the running time on smaller graphs by up to a factor of 2x. Our SCC implementation is between 1.6x faster to 4.8x slower than running connectivity on the graph. There are several existing SCC implementations that have been evaluated on real-world directed graphs (61, 121, 85). The Hong et al. algorithm (61) is a modified version of the FWBW-Trim algorithm from McLendon et al. (85), but neither algorithm has any theoretical bounds on work or depth. Unfortunately (61) do not report running times, so we are unable to compare our performance with them. The Multistep algorithm (121) has a worst-case running time of , but the authors point-out that the algorithm behaves like a linear-time algorithm on real-world graphs. We ran our implementation on 16 cores configured similarly to their experiments and found that we are about 1.7x slower on LiveJournal, which easily fits in cache, and 1.2x faster on Twitter (scaled to account for a small difference in graph sizes). While the multistep algorithm is slightly faster on some graphs, our SCC implementation has the advantage of being theoretically-efficient and performs a predictable amount of work.
Our minimum spanning forest implementation achieves between 17–54x speedup over the implementation running on a single thread across all of our inputs. Obtaining practical parallel algorithms for MSF has been a longstanding goal in the field, and several existing implementations exist (9, 94, 36, 116, 130). We compared our implementation with the union-find based MSF implementation from PBBS (116) and the implementation of Borůvka from (130), which is one of the fastest implementations we are aware of. Our MSF implementation is between 2.6–5.9x faster than the MSF implementation from PBBS. Compared to the edgelist based implementation of Borůvka from (130) our implementation is between 1.2–2.9x faster.
7.4. Covering Problems
Our MIS and maximal matching implementations achieve between 31–70x and 25–70x speedup across all inputs. The implementations by Blelloch et al. (22) are the fastest existing implementations of MIS and maximal matching that we are aware of, and are the basis for our maximal matching implementation. They report that their implementations are 3–8x faster than Luby’s algorithm on 32 threads, and outperform a sequential greedy MIS implementation on more than 2 processors. We compared our rootset-based MIS implementation to the prefix-based implementation, and found that the rootset-based approach is between 1.1–3.5x faster. Our maximal matching implementation is between 3–4.2x faster than the implementation from (22). Our implementation of maximal matching can avoid a significant amount of work, as each of the filter steps can extract and permute just the highest priority edges, whereas the edgelist-based version in PBBS must permute all edges. Our coloring implementation achieves between 11–56x speedup across all inputs. We note that our implementation appears to be between 1.2–1.6x slower than the asynchronous implementation of JP in (59), due to synchronizing on many rounds which contain few vertices.
Our approximate set cover implementation achieves between 5–57x speedup across all inputs. Our implementation is based on the implementation presented in Julienne (42); the one major modification was to regenerate random priorities for sets that are active on the current round. We compared the running time of our implementation with the parallel implementation from (26) which is available in the PBBS library. We ran both implementations with . Our implementation is between 1.2x slower to 1.5x faster than the PBBS implementation on our graphs, with the exception of 3D-Torus. On 3D-Torus, the implementation from (26) runs 56x slower than our implementation as it does not regenerate priorities for active sets on each round causing worst-case behavior. Our performance is also slow on this graph, as nearly all of the vertices stay active (in the highest bucket) during each round, and using causes a large number of rounds to be performed.
7.5. Substructure Problems
Our -core implementation achieves between 5–46x speedup across all inputs, and 114x speedup on the 3D-Torus graph as there is only one round of peeling in which all vertices are removed. There are several recent papers that implement parallel algorithms for -core (40, 42, 107, 68). Both the ParK algorithm (40) and Kabir and Madduri algorithm (68) implement the peeling algorithm in work, which is not work-efficient. Our implementation is between 3.8–4.6x faster than ParK on a similar machine configuration. Kabir and Madduri show that their implementation achieves an average speedup of 2.8x over ParK. Our implementation is between 1.3–1.6x faster than theirs on a similar machine configuration.
Our approximate densest subgraph implementation achieves between 44–77x speedup across all inputs. We ran our implementation with , which in our experiments produced subgraphs with density roughly equal to those produced by the 2-approximation algorithm based on degeneracy ordering, or setting to 0. To the best of our knowledge, there are no prior existing shared-memory parallel algorithms for this problem.
Our triangle counting (TC) implementation achieves between 39–81x speedup across all inputs. Unfortunately, we are unable to report speedup numbers for TC on our larger graphs as the single-threaded times took too long due to the algorithm performing work. There are a number of experimental papers that consider multicore triangle counting (112, 54, 72, 119, 1, 77). We implement the algorithm from (119), and adapted it to work on compressed graphs. We note that in our experiments we intersect directed adjacency lists sequentially, as there was sufficient parallelism in the outer parallel-loop. There was no significant difference in running times between our implementation and the implementation from (119). We ran our implementation on 48 threads on the Twitter graph to compare with the times reported by EmptyHeaded (1) and found that our times are about the same.
7.6. Eigenvector Problems
Our PageRank (PR) implementation achieves between 39–54x speedup across all inputs. Our implementation is based on the PageRank-Delta implementation from Ligra (114). We ran the algorithm with , and with , where is the local (per-node) threshold for the amount of change in its PageRank value before it is activated. We note that the modification made to carry out dense iterations using a reduction over the in-neighbors of a vertex was important to decrease contention, and provided between 2–3x speedup over the Ligra implementation in practice. Many graph processing systems implement PageRank. The optimizing compiler used by GraphIt generates a highly-optimized implementation that is currently the fastest shared-memory implementation known to us (129). We note that our implementation is about 1.8x slower than the implementation in GraphIt for LiveJournal and Twitter when run on the same number of threads as in their experiments.
7.7. Performance on 3D-Torus
We ran experiments on a family of 3D-Torus graphs with different sizes to study how our diameter-bounded algorithms scale relative to algorithms with polylogarithmic depth. We were surprised to see that the running time of some of our polylogarithmic depth algorithms on this graph, like LDD and connectivity, are 17–40x more expensive than their running time on Twitter and Twitter-Sym, despite 3D-Torus only having 4x and 2.4x more edges than Twitter and Twitter-Sym. Our slightly worse scaling on this graph can be accounted for by the fact that we stored the graph ordered by dimension, instead of storing it using a local ordering. It would be interesting to see how much improvement we could gain by reordering the vertices.
In Figure 1 we show the normalized throughput of MIS, BFS, BC, and graph coloring for 3-dimensional tori of different sizes, where throughput is measured as the number of edges processed per second. The throughput for each application becomes saturated before our largest-scale graph for all applications except for BFS, which is saturated on a graph with 2 billion vertices. The throughput curves show that the theoretical bounds are useful in predicting how the half-lengths55 5 The graph size when the system achieves half of its peak-performance. are distributed. The half-lengths are ordered as follows: coloring, MIS, BFS, and BC. This is the same order as sorting these algorithms by their depth with respect to this graph.
Locality. While our algorithms are efficient on the , we do not analyze their cache complexity, and in general they may not be efficient in a model that takes caches into account. Despite this, we observed that our algorithms have good cache performance on the graphs we tested on. In this section we give some explanation for this fact by showing that our primitives make good use of the caches. Our algorithms are also aided by the fact that these graph datasets often come in highly local orders (e.g., see the Natural order in (43)). Table 5 shows metrics for our experiments measured using Open Performance Counter Monitor (PCM).
| Algorithm | Cycles Stalled | LLC Hit Rate | LLC Misses | BW | Time |
|---|---|---|---|---|---|
| -core (histogram) | 9 | 0.223 | 49 | 96 | 62.9 |
| -core (fetch-and-add) | 67 | 0.155 | 42 | 24 | 221 |
| weighted BFS (blocked) | 3.7 | 0.070 | 19 | 130 | 14.4 |
| weighted BFS (unblocked) | 5.6 | 0.047 | 29 | 152 | 25.2 |
Due to space limitations, we only report numbers for the ClueWeb graph. We observed that using a work-efficient histogram is 3.5x faster than using fetch-and-add in our -core implementation, which suffers from high contention on this graph. Using a histogram reduces the number of cycles stalled due to memory by more than 7x. We also ran our wBFS implementation with and without the edgeMapBlocked optimization, which reduces the number of cache-lines read from and written to when performing a sparse edgeMap. The blocked implementation reads and writes 2.1x fewer bytes than the unoptimized version, which translates to a 1.7x faster runtime. We disabled the dense optimization for this experiment to directly compare the two implementations of a sparse edgeMap.
7.8. Processing Massive Web Graphs
In Tables 3 and 4, we show the running times of our implementations on the ClueWeb, Hyperlink2014, and Hyperlink2012 graphs. To put our performance in context, we compare our 72-core running times to running times reported by existing work. Table 6 summarizes state-of-the-art existing results in the literature. Most results process the directed versions of these graphs, which have about half as many edges as the symmetrized version. Unless otherwise mentioned, all results from the literature use the directed versions of these graphs. To make the comparison easier we show our running times for BFS, SSSP (weighted BFS), BC and SCC on the directed graphs, and running times for Connectivity, -core and TC on the symmetrized graphs in Table 6.
| Paper | Problem | Graph | Memory | Hyper-threads | Nodes | Time |
| Mosaic (79) | BFS* | 2014 | 0.768 | 1000 | 1 | 6.55 |
| Connectivity* | 2014 | 0.768 | 1000 | 1 | 708 | |
| SSSP* | 2014 | 0.768 | 1000 | 1 | 8.6 | |
| FlashGraph (39) | BFS* | 2012 | .512 | 64 | 1 | 208 |
| BC* | 2012 | .512 | 64 | 1 | 595 | |
| Connectivity* | 2012 | .512 | 64 | 1 | 461 | |
| TC* | 2012 | .512 | 64 | 1 | 7818 | |
| BigSparse (67) | BFS* | 2012 | 0.064 | 32 | 1 | 2500 |
| BC* | 2012 | 0.064 | 32 | 1 | 3100 | |
| Slota et al. (123) | Largest-CC* | 2012 | 16.3 | 8192 | 256 | 63 |
| Largest-SCC* | 2012 | 16.3 | 8192 | 256 | 108 | |
| Approx -core* | 2012 | 16.3 | 8192 | 256 | 363 | |
| Stergiou et al. (124) | Connectivity | 2012 | 128 | 24000 | 1000 | 341 |
| Gluon (41) | BFS | 2012 | 24 | 69632 | 256 | 380 |
| Connectivity | 2012 | 24 | 69632 | 256 | 75.3 | |
| PageRank | 2012 | 24 | 69632 | 256 | 158.2 | |
| SSSP | 2012 | 24 | 69632 | 256 | 574.9 | |
| This paper | BFS* | 2014 | 1 | 144 | 1 | 5.71 |
| SSSP* | 2014 | 1 | 144 | 1 | 9.08 | |
| Connectivity | 2014 | 1 | 144 | 1 | 11.2 | |
| BFS* | 2012 | 1 | 144 | 1 | 16.7 | |
| BC* | 2012 | 1 | 144 | 1 | 35.2 | |
| Connectivity | 2012 | 1 | 144 | 1 | 25.0 | |
| SCC* | 2012 | 1 | 144 | 1 | 185 | |
| SSSP | 2012 | 1 | 144 | 1 | 58.1 | |
| -core | 2012 | 1 | 144 | 1 | 184 | |
| PageRank | 2012 | 1 | 144 | 1 | 462 | |
| TC | 2012 | 1 | 144 | 1 | 1168 |
FlashGraph (39) reports disk-based running times for the Hyperlink2012 graph on a 4-socket, 32-core machine with 512GB of memory and 15 SSDs. On 64 hyper-threads, they solve BFS in 208s, BC in 595s, connected components in 461s, and triangle counting in 7818s. Our BFS and BC implementations are 12x faster and 16x faster, and our triangle counting and connectivity implementations are 5.3x faster and 18x faster than their implementations, respectively. Mosaic (79) report in-memory running times on the Hyperlink2014 graph; we note that the system is optimized for external memory execution. They solve BFS in 6.5s, connected components in 700s, and SSSP (Bellman-Ford) in 8.6s on a machine with 24 hyper-threads and 4 Xeon-Phis (244 cores with 4 threads each) for a total of 1000 hyper-threads, 768GB of RAM, and 6 NVMes. Our BFS and connectivity implementations are 1.1x and 62x faster respectively, and our SSSP implementation is 1.05x slower. Both FlashGraph and Mosaic compute weakly connected components, which is equivalent to connectivity. BigSparse (67) report disk-based running times for BFS and BC on the Hyperlink2012 graph on a 32-core machine. They solve BFS in 2500s and BC in 3100s. Our BFS and BC implementations are 149x and 88x faster than their implementations, respectively.
Slota et al. (123) report running times for the Hyperlink2012 graph on 256 nodes on the Blue Waters supercomputer. Each node contains two 16-core processors with one thread each, for a total of 8192 hyper-threads. They report they can find the largest connected component and SCC from the graph in 63s and 108s respectively. Our implementations find all connected components 2.5x faster than their largest connected component implementation, and find all strongly connected components 1.6x slower than their largest-SCC implementation. Their largest-SCC implementation computes two BFSs from a randomly chosen vertex—one on the in-edges and the other on the out-edges—and intersects the reachable sets. We perform the same operation as one of the first steps of our SCC algorithm and note that it requires about 30 seconds on our machine. They solve approximate -cores in 363s, where the approximate -core of a vertex is the coreness of the vertex rounded up to the nearest powers of 2. Our implementation computes the exact coreness of each vertex in 184s, which is 1.9x faster than the approximate implementation while using 113x fewer cores.
Recently, Dathathri et al. (41) have reported running times for the Hyperlink2012 graph using Gluon, a distributed graph processing system based on Galois. They process this graph on a 256 node system, where each node is equipped with 68 4-way hyper-threaded cores, and the hosts are connected by an Intel Omni-Path network with 100Gbps peak bandwidth. They report times for BFS, connectivity, PageRank, and SSSP. Other than their connectivity implementation, which uses pointer-jumping, their implementations are based on data-driven asynchronous label-propagation. We are not aware of any theoretical bounds on the work and depth of these implementations. Compared to their reported times, our implementation of BFS is 22.7x faster, our implementation of connectivity is 3x faster, and our implementation of SSSP is 9.8x faster. Our PageRank implementation is 2.9x slower. However, we note that the PageRank numbers they report are not for true PageRank, but PageRank-Delta, and are thus in some sense incomparable.
Stergiou et al. (124) describe a connectivity algorithm that runs in rounds in the BSP model and report running times for the symmetrized Hyperlink2012 graph. They implement their algorithm using a proprietary in-memory/secondary-storage graph processing system used at Yahoo!, and run experiments on a 1000 node cluster. Each node contains two 6-core processors that are 2-way hyper-threaded and 128GB of RAM, for a total of 24000 hyper-threads and 128TB of RAM. Their fastest running time on the Hyperlink2012 graph is 341s on their 1000 node system. Our implementation solves connectivity on this graph in 25s–13.6x faster on a system with 128x less memory and 166x fewer cores. They also report running times for solving connectivity on a private Yahoo! webgraph with 272 billion vertices and 5.9 trillion edges, over 26 times the size of our largest graph. While such a graph seems to currently be out of reach of our machine, we are hopeful that techniques from theoretically-efficient parallel algorithms can help solve problems on graphs at this scale and beyond.
8. Conclusion
In this paper, we showed that we can process the largest publicly-available real-world graph on a single shared-memory server with 1TB of memory using theoretically-efficient parallel algorithms. We outperform existing implementations on the largest real-world graphs, and use much fewer resources than the distributed-memory solutions. On a per-core basis, our numbers are significantly better. Our results provide evidence that theoretically-efficient shared-memory graph algorithms can be efficient and scalable in practice.
Acknowledgements
Thanks to the reviewers and to Lin Ma for helpful comments. This research was supported in part by NSF grants #CCF-1408940, #CCF-1533858, #CCF-1629444, and #CCF-1845763, and DOE grant #DE-SC0018947.
References
- [1] C. R. Aberger, A. Lamb, S. Tu, A. Nötzli, K. Olukotun, and C. Ré. Emptyheaded: A relational engine for graph processing. ACM Trans. Database Syst., 2017.
- [2] A. Aggarwal, R. J. Anderson, and M.-Y. Kao. Parallel depth-first search in general directed graphs. In STOC, 1989.
- [3] M. Ahmad, F. Hijaz, Q. Shi, and O. Khan. Crono: A benchmark suite for multithreaded graph algorithms executing on futuristic multicores. In IISWC, 2015.
- [4] N. Alon, L. Babai, and A. Itai. A fast and simple randomized parallel algorithm for the maximal independent set problem. J. Algorithms, 1986.
- [5] N. Alon, R. Yuster, and U. Zwick. Finding and counting given length cycles. Algorithmica, 17(3), 1997.
- [6] R. Anderson and E. W. Mayr. A -complete problem and approximations to it. Technical report, 1984.
- [7] B. Awerbuch. Complexity of network synchronization. J. ACM, 32(4), 1985.
- [8] B. Awerbuch and Y. Shiloach. New connectivity and MSF algorithms for Ultracomputer and PRAM. In ICPP, 1983.
- [9] D. A. Bader and G. Cong. Fast shared-memory algorithms for computing the minimum spanning forest of sparse graphs. JPDC, 2006.
- [10] D. A. Bader and K. Madduri. Design and implementation of the hpcs graph analysis benchmark on symmetric multiprocessors. In International Conference on High-Performance Computing, 2005.
- [11] D. A. Bader and K. Madduri. Designing multithreaded algorithms for breadth-first search and st-connectivity on the cray mta-2. In ICPP, 2006.
- [12] B. Bahmani, A. Goel, and K. Munagala. Efficient primal-dual graph algorithms for mapreduce. In International Workshop on Algorithms and Models for the Web-Graph, pages 59–78, 2014.
- [13] B. Bahmani, R. Kumar, and S. Vassilvitskii. Densest subgraph in streaming and mapreduce. VLDB, 5(5):454–465, 2012.
- [14] G. Baier, E. Köhler, and M. Skutella. The k-splittable flow problem. Algorithmica, 42(3-4):231–248, 2005.
- [15] S. Beamer, K. Asanovic, and D. A. Patterson. The GAP benchmark suite. CoRR, abs/1508.03619, 2015.
- [16] N. Ben-David, G. E. Blelloch, J. T. Fineman, P. B. Gibbons, Y. Gu, C. McGuffey, and J. Shun. Implicit decomposition for write-efficient connectivity algorithms. In IPDPS, 2018.
- [17] B. Berger, J. Rompel, and P. W. Shor. Efficient algorithms for set cover with applications to learning and geometry. J. Comput. Syst. Sci., 49(3), Dec. 1994.
- [18] M. Birn, V. Osipov, P. Sanders, C. Schulz, and N. Sitchinava. Efficient parallel and external matching. In Euro-Par, 2013.
- [19] G. E. Blelloch. Prefix sums and their applications. Synthesis of Parallel Algorithms, 1993.
- [20] G. E. Blelloch and L. Dhulipala. Introduction to parallel algorithms 15-853: Algorithms in the real world. 2018.
- [21] G. E. Blelloch, J. T. Fineman, P. B. Gibbons, and J. Shun. Internally deterministic algorithms can be fast. In PPoPP, 2012.
- [22] G. E. Blelloch, J. T. Fineman, and J. Shun. Greedy sequential maximal independent set and matching are parallel on average. In SPAA, 2012.
- [23] G. E. Blelloch, Y. Gu, J. Shun, and Y. Sun. Parallelism in randomized incremental algorithms. In SPAA, 2016.
- [24] G. E. Blelloch, Y. Gu, and Y. Sun. A new efficient construction on probabilistic tree embeddings. In ICALP, 2017.
- [25] G. E. Blelloch, R. Peng, and K. Tangwongsan. Linear-work greedy parallel approximate set cover and variants. In SPAA, 2011.
- [26] G. E. Blelloch, H. V. Simhadri, and K. Tangwongsan. Parallel and I/O efficient set covering algorithms. In SPAA, 2012.
- [27] R. D. Blumofe and C. E. Leiserson. Scheduling multithreaded computations by work stealing. J. ACM, 46(5), Sept. 1999.
- [28] P. Boldi and S. Vigna. The WebGraph framework I: Compression techniques. In WWW, 2004.
- [29] O. Borůvka. O jistém problému minimálním. Práce Mor. Přírodověd. Spol. v Brně III, 3, 1926.
- [30] U. Brandes. A faster algorithm for betweenness centrality. Journal of mathematical sociology, 25(2), 2001.
- [31] S. Brin and L. Page. The anatomy of a large-scale hypertextual web search engine. In WWW, pages 107–117, 1998.
- [32] A. Broder, R. Kumar, F. Maghoul, P. Raghavan, S. Rajagopalan, R. Stata, A. Tomkins, and J. Wiener. Graph structure in the web. Computer networks, 33(1-6), 2000.
- [33] M. Charikar. Greedy approximation algorithms for finding dense components in a graph. APPROX, pages 84–95, 2000.
- [34] R. Cole, P. N. Klein, and R. E. Tarjan. Finding minimum spanning forests in logarithmic time and linear work using random sampling. In SPAA, 1996.
- [35] G. Cong and D. A. Bader. An experimental study of parallel biconnected components algorithms on symmetric multiprocessors (SMPs). In IPDPS, 2005.
- [36] G. Cong and I. G. Tanase. Composable locality optimizations for accelerating parallel forest computations. In HPCC, 2016.
- [37] D. Coppersmith, L. Fleischer, B. Hendrickson, and A. Pinar. A divide-and-conquer algorithm for identifying strongly connected components. 2003.
- [38] T. H. Cormen, C. E. Leiserson, R. L. Rivest, and C. Stein. Introduction to Algorithms (3. ed.). MIT Press, 2009.
- [39] D. M. Da Zheng, R. Burns, J. Vogelstein, C. E. Priebe, and A. S. Szalay. Flashgraph: Processing billion-node graphs on an array of commodity SSDs. In FAST, 2015.
- [40] N. S. Dasari, R. Desh, and M. Zubair. ParK: An efficient algorithm for -core decomposition on multicore processors. In Big Data, 2014.
- [41] R. Dathathri, G. Gill, L. Hoang, H.-V. Dang, A. Brooks, N. Dryden, M. Snir, and K. Pingali. Gluon: A communication-optimizing substrate for distributed heterogeneous graph analytics. In PLDI, pages 752–768. ACM, 2018.
- [42] L. Dhulipala, G. E. Blelloch, and J. Shun. Julienne: A framework for parallel graph algorithms using work-efficient bucketing. In SPAA, 2017.
- [43] L. Dhulipala, I. Kabiljo, B. Karrer, G. Ottaviano, S. Pupyrev, and A. Shalita. Compressing graphs and indexes with recursive graph bisection. In KDD, 2016.
- [44] R. Duan, K. Lyu, and Y. Xie. Single-source bottleneck path algorithm faster than sorting for sparse graphs. In ICALP, pages 43:1–43:14, 2018.
- [45] J. A. Edwards and U. Vishkin. Better speedups using simpler parallel programming for graph connectivity and biconnectivity. In PMAM, 2012.
- [46] J. T. Fineman. Nearly work-efficient parallel algorithm for digraph reachability. In STOC, 2018.
- [47] M. Fischer and A. Noever. Tight analysis of parallel randomized greedy MIS. In SODA, 2018.
- [48] L. K. Fleischer, B. Hendrickson, and A. Pinar. On identifying strongly connected components in parallel. In IPDPS, 2000.
- [49] L. R. Ford and D. R. Fulkerson. Maximal flow through a network. In Classic papers in combinatorics, pages 243–248. Springer, 2009.
- [50] H. Gazit. An optimal randomized parallel algorithm for finding connected components in a graph. SIAM J. Comput., 1991.
- [51] H. Gazit and G. L. Miller. An improved parallel algorithm that computes the BFS numbering of a directed graph. Information Processing Letters, 28(2), 1988.
- [52] A. V. Goldberg. Finding a maximum density subgraph. Technical report, Berkeley, CA, USA, 1984.
- [53] J. E. Gonzalez, Y. Low, H. Gu, D. Bickson, and C. Guestrin. PowerGraph: Distributed graph-parallel computation on natural graphs. In OSDI, 2012.
- [54] O. Green, L. M. Munguia, and D. A. Bader. Load balanced clustering coefficients. In PPAA, 2014.
- [55] R. Greenlaw, H. J. Hoover, and W. L. Ruzzo. Limits to Parallel Computation: P-completeness Theory. Oxford University Press, Inc., 1995.
- [56] Y. Gu, J. Shun, Y. Sun, and G. E. Blelloch. A top-down parallel semisort. In SPAA, 2015.
- [57] S. Halperin and U. Zwick. An optimal randomized logarithmic time connectivity algorithm for the EREW PRAM (extended abstract). In SPAA, 1994.
- [58] S. Halperin and U. Zwick. Optimal randomized EREW PRAM algorithms for finding spanning forests. In J. Algorithms, 2000.
- [59] W. Hasenplaugh, T. Kaler, T. B. Schardl, and C. E. Leiserson. Ordering heuristics for parallel graph coloring. In SPAA, 2014.
- [60] L. Hoang, M. Pontecorvi, R. Dathathri, G. Gill, B. You, K. Pingali, and V. Ramachandran. A round-efficient distributed betweenness centrality algorithm. In PPoPP, pages 272–286. ACM, 2019.
- [61] S. Hong, N. C. Rodia, and K. Olukotun. On fast parallel detection of strongly connected components (SCC) in small-world graphs. In SC, 2013.
- [62] J. Hopcroft and R. Tarjan. Algorithm 447: efficient algorithms for graph manipulation. Communications of the ACM, 1973.
- [63] A. Iosup, T. Hegeman, W. L. Ngai, S. Heldens, A. Prat-Pérez, T. Manhardto, H. Chafio, M. Capotă, N. Sundaram, M. Anderson, I. G. Tănase, Y. Xia, L. Nai, and P. Boncz. LDBC graphalytics: A benchmark for large-scale graph analysis on parallel and distributed platforms. Proc. VLDB Endow., 9(13), Sept. 2016.
- [64] A. Israeli and Y. Shiloach. An improved parallel algorithm for maximal matching. Inf. Process. Lett., 1986.
- [65] A. Itai and M. Rodeh. Finding a minimum circuit in a graph. In STOC, 1977.
- [66] J. Jaja. Introduction to Parallel Algorithms. Addison-Wesley Professional, 1992.
- [67] S. W. Jun, A. Wright, S. Zhang, S. Xu, and Arvind. BigSparse: High-performance external graph analytics. CoRR, abs/1710.07736, 2017.
- [68] H. Kabir and K. Madduri. Parallel k-core decomposition on multicore platforms. In IPDPSW, 2017.
- [69] D. R. Karger, P. N. Klein, and R. E. Tarjan. A randomized linear-time algorithm to find minimum spanning trees. J. ACM, 42(2), Mar. 1995.
- [70] R. M. Karp and V. Ramachandran. Handbook of theoretical computer science (vol. a). chapter Parallel Algorithms for Shared-memory Machines. MIT Press, Cambridge, MA, USA, 1990.
- [71] R. M. Karp and A. Wigderson. A fast parallel algorithm for the maximal independent set problem. In STOC, 1984.
- [72] J. Kim, W.-S. Han, S. Lee, K. Park, and H. Yu. OPT: A new framework for overlapped and parallel triangulation in large-scale graphs. In SIGMOD, 2014.
- [73] R. Kumar, B. Moseley, S. Vassilvitskii, and A. Vattani. Fast greedy algorithms in mapreduce and streaming. ACM Trans. Parallel Comput., 2(3), Sept. 2015.
- [74] H. Kwak, C. Lee, H. Park, and S. Moon. What is twitter, a social network or a news media? In WWW, 2010.
- [75] M. Latapy. Main-memory triangle computations for very large (sparse (power-law)) graphs. Theor. Comput. Sci., 2008.
- [76] C. E. Leiserson and T. B. Schardl. A work-efficient parallel breadth-first search algorithm (or how to cope with the nondeterminism of reducers). In SPAA, 2010.
- [77] Y. Low, J. Gonzalez, A. Kyrola, D. Bickson, C. Guestrin, and J. M. Hellerstein. GraphLab: A new parallel framework for machine learning. In UAI, 2010.
- [78] M. Luby. A simple parallel algorithm for the maximal independent set problem. SIAM J. Comput., 1986.
- [79] S. Maass, C. Min, S. Kashyap, W. Kang, M. Kumar, and T. Kim. Mosaic: Processing a trillion-edge graph on a single machine. In EuroSys, 2017.
- [80] S. Maleki, D. Nguyen, A. Lenharth, M. Garzarán, D. Padua, and K. Pingali. DSMR: A parallel algorithm for single-source shortest path problem. In ICS, 2016.
- [81] G. Malewicz, M. H. Austern, A. J. Bik, J. C. Dehnert, I. Horn, N. Leiser, and G. Czajkowski. Pregel: A system for large-scale graph processing. In SIGMOD, 2010.
- [82] Y. Maon, B. Schieber, and U. Vishkin. Parallel ear decomposition search (EDS) and st-numbering in graphs. Theoretical Computer Science, 47, 1986.
- [83] D. W. Matula and L. L. Beck. Smallest-last ordering and clustering and graph coloring algorithms. J. ACM, 30(3), July 1983.
- [84] R. R. McCune, T. Weninger, and G. Madey. Thinking like a vertex: A survey of vertex-centric frameworks for large-scale distributed graph processing. ACM Comput. Surv., 48(2), Oct. 2015.
- [85] W. Mclendon Iii, B. Hendrickson, S. J. Plimpton, and L. Rauchwerger. Finding strongly connected components in distributed graphs. Journal of Parallel and Distributed Computing, 65(8), 2005.
- [86] R. Meusel, S. Vigna, O. Lehmberg, and C. Bizer. The graph structure in the web–analyzed on different aggregation levels. The Journal of Web Science, 1(1), 2015.
- [87] U. Meyer and P. Sanders. Parallel shortest path for arbitrary graphs. In Euro-Par, 2000.
- [88] U. Meyer and P. Sanders. -stepping: a parallelizable shortest path algorithm. J. Algorithms, 49(1), 2003.
- [89] G. L. Miller, R. Peng, A. Vladu, and S. C. Xu. Improved parallel algorithms for spanners and hopsets. pages 192–201, 2015.
- [90] G. L. Miller, R. Peng, and S. C. Xu. Parallel graph decompositions using random shifts. In SPAA, 2013.
- [91] G. L. Miller and V. Ramachandran. A new graph triconnectivity algorithm and its parallelization. Combinatorica, 12(1), Mar 1992.
- [92] L. Nai, Y. Xia, I. G. Tanase, H. Kim, and C.-Y. Lin. GraphBIG: Understanding graph computing in the context of industrial solutions. In SC, 2015.
- [93] D. Nguyen, A. Lenharth, and K. Pingali. A lightweight infrastructure for graph analytics. In SOSP, 2013.
- [94] S. Nobari, T.-T. Cao, P. Karras, and S. Bressan. Scalable parallel minimum spanning forest computation. In PPoPP, 2012.
- [95] M. Ortmann and U. Brandes. Triangle listing algorithms: Back from the diversion. In ALENEX, 2014.
- [96] V. Osipov, P. Sanders, and J. Singler. The filter-kruskal minimum spanning tree algorithm. In ALENEX, 2009.
- [97] R. Pagh and F. Silvestri. The input/output complexity of triangle enumeration. In PODS, 2014.
- [98] M. Patwary, P. Refsnes, and F. Manne. Multi-core spanning forest algorithms using the disjoint-set data structure. In IPDPS, 2012.
- [99] D. Peleg and A. A. Schäffer. Graph spanners. Journal of graph theory, 13(1):99–116, 1989.
- [100] S. Pettie and V. Ramachandran. A randomized time-work optimal parallel algorithm for finding a minimum spanning forest. SIAM J. Comput., 31(6), 2002.
- [101] C. A. Phillips. Parallel graph contraction. In SPAA, 1989.
- [102] C. K. Poon and V. Ramachandran. A randomized linear work EREW PRAM algorithm to find a minimum spanning forest. In ISAAC, 1997.
- [103] S. Rajagopalan and V. V. Vazirani. Primal-dual approximation algorithms for set cover and covering integer programs. SIAM J. Comput., 28(2), Feb. 1999.
- [104] V. Ramachandran. A framework for parallel graph algorithm design. In Optimal Algorithms, 1989.
- [105] V. Ramachandran. Parallel open ear decomposition with applications to graph biconnectivity and triconnectivity. In Synthesis of Parallel Algorithms, 1993.
- [106] J. Reif. Optimal parallel algorithms for integer sorting and graph connectivity. TR-08-85, Harvard University, 1985.
- [107] A. E. Sariyuce, C. Seshadhri, and A. Pinar. Parallel local algorithms for core, truss, and nucleus decompositions. CoRR, abs/1704.00386, 2017.
- [108] T. Schank. Algorithmic aspects of triangle-based network analysis. PhD Thesis, Universitat Karlsruhe, 2007.
- [109] T. Schank and D. Wagner. Finding, counting and listing all triangles in large graphs, an experimental study. In WEA, 2005.
- [110] W. Schudy. Finding strongly connected components in parallel using reachability queries. In SPAA, 2008.
- [111] S. B. Seidman. Network structure and minimum degree. Soc. Networks, 5(3), 1983.
- [112] M. Sevenich, S. Hong, A. Welc, and H. Chafi. Fast in-memory triangle listing for large real-world graphs. In Workshop on Social Network Mining and Analysis, 2014.
- [113] Y. Shiloach and U. Vishkin. An parallel connectivity algorithm. J. Algorithms, 1982.
- [114] J. Shun and G. E. Blelloch. Ligra: A lightweight graph processing framework for shared memory. In PPoPP, 2013.
- [115] J. Shun, G. E. Blelloch, J. T. Fineman, and P. B. Gibbons. Reducing contention through priority updates. In SPAA, 2013.
- [116] J. Shun, G. E. Blelloch, J. T. Fineman, P. B. Gibbons, A. Kyrola, H. V. Simhadri, and K. Tangwongsan. Brief announcement: the Problem Based Benchmark Suite. In SPAA, 2012.
- [117] J. Shun, L. Dhulipala, and G. E. Blelloch. A simple and practical linear-work parallel algorithm for connectivity. In SPAA, 2014.
- [118] J. Shun, L. Dhulipala, and G. E. Blelloch. Smaller and faster: Parallel processing of compressed graphs with Ligra+. In DCC, 2015.
- [119] J. Shun and K. Tangwongsan. Multicore triangle computations without tuning. In ICDE, 2015.
- [120] G. M. Slota and K. Madduri. Simple parallel biconnectivity algorithms for multicore platforms. In HiPC, 2014.
- [121] G. M. Slota, S. Rajamanickam, and K. Madduri. BFS and coloring-based parallel algorithms for strongly connected components and related problems. In IPDPS, 2014.
- [122] G. M. Slota, S. Rajamanickam, and K. Madduri. Supercomputing for Web Graph Analytics. Apr 2015.
- [123] G. M. Slota, S. Rajamanickam, and K. Madduri. A case study of complex graph analysis in distributed memory: Implementation and optimization. In IPDPS, 2016.
- [124] S. Stergiou, D. Rughwani, and K. Tsioutsiouliklis. Shortcutting label propagation for distributed connected components. In WSDM, 2018.
- [125] R. E. Tarjan and U. Vishkin. An efficient parallel biconnectivity algorithm. SIAM Journal on Computing, 1985.
- [126] M. Thorup and U. Zwick. Approximate distance oracles. Journal of the ACM (JACM), 52(1):1–24, 2005.
- [127] D. J. Welsh and M. B. Powell. An upper bound for the chromatic number of a graph and its application to timetabling problems. The Computer Journal, 1967.
- [128] D. Yan, Y. Bu, Y. Tian, and A. Deshpande. Big graph analytics platforms. Foundations and Trends in Databases, 7, 2017.
- [129] Y. Zhang, M. Yang, R. Baghdadi, S. Kamil, J. Shun, and S. Amarasinghe. Graphit: A high-performance graph dsl. Proc. ACM Program. Lang., 2(OOPSLA):121:1–121:30, Oct. 2018.
- [130] W. Zhou. A practical scalable shared-memory parallel algorithm for computing minimum spanning trees. Master’s thesis, KIT, 2017.
Appendix A Graph Statistics
In this section, we list graph statistics computed for the graphs from Section 7.66 6 Similar statistics can be found on the SNAP website (https://snap.stanford.edu/data/) and the Laboratory for Web Algorithmics website (http://law.di.unimi.it/datasets.php). These statistics include the number of connected components, strongly connected components, colors used by the LLF and LF heuristics, number of triangles, and several others. These numbers will be useful for verifying the correctness or quality of our algorithms in relation to future algorithms that also run on these graphs. Although some of these numbers were present in Table 2, we include in the tables below for completeness.
To ensure that the interested reader can reproduce these results we provide details about the statistics that are not self-explanatory.
- •
Effective Directed Diameter: the maximum number of levels traversed during a graph traversal algorithm (BFS or SCC) on the unweighted directed graph.
- •
Effective Undirected Diameter: the maximum number of levels traversed during a graph traversal algorithm (BFS) on the unweighted directed graph.
- •
Size of Largest (Connected/Biconnected/Strongly-Connected) Component: The number of vertices in the largest (connected/biconnected/strongly-connected) component. Note that in the case of biconnectivity, we assign labels to edges, so a vertex participates in a component for each distinct edge label incident to it.
- •
Num. Triangles: The number of closed triangles in , where each triangle is counted exactly once.
- •
Num. Colors Used by (LF/LLF): The number of colors used is just the maximum color ID assigned to any vertex.
- •
(Maximum Independent Set/Maximum Matching/Approximate Set Cover) Size: We report the sizes of these objects computed by our implementations. For MIS and maximum matching we report this metric to lower-bound on the size of the maximum independent set and maximum matching supported by the graph. For approximate set cover, we run our code on instances similar to those used in prior work (e.g., Blelloch et al. [26] and Dhulipala et al. [42]) where the elements are vertices and the sets are the neighbors of each vertex in the undirected graph. In the case of the social network and hyperlink graphs, this optimization problem naturally captures the minimum number of users or Web pages whose neighborhoods must be retrieved to cover the entire graph.
- •
(Degeneracy): The value of of the largest non-empty -core.
| Statistic | Value |
|---|---|
| Num. Vertices | 4,847,571 |
| Num. Directed Edges | 68,993,773 |
| Num. Undirected Edges | 85,702,474 |
| Effective Directed Diameter | 16 |
| Effective Undirected Diameter | 20 |
| Num. Connected Components | 1,876 |
| Num. Biconnected Components | 1,133,883 |
| Num. Strongly Connected Components | 971,232 |
| Size of Largest Connected Component | 4,843,953 |
| Size of Largest Biconnected Component | 3,665,291 |
| Size of Largest Strongly Connected Component | 3,828,682 |
| Num. Triangles | 285,730,264 |
| Num. Colors Used by LF | 323 |
| Num. Colors Used by LLF | 327 |
| Maximal Independent Set Size | 2,316,617 |
| Maximal Matching Size | 1,546,833 |
| Set Cover Size | 964,492 |
| (Degeneracy) | 372 |
| (Num. Peeling Rounds in -core) | 3,480 |
| Statistic | Value |
| Num. Vertices | 3,072,627 |
| Num. Directed Edges | — |
| Num. Undirected Edges | 234,370,166 |
| Effective Directed Diameter | — |
| Effective Undirected Diameter | 9 |
| Num. Connected Components | 187 |
| Num. Biconnected Components | 68,117 |
| Num. Strongly Connected Components | — |
| Size of Largest Connected Component | 3,072,441 |
| Size of Largest Biconnected Component | 3,003,914 |
| Size of Largest Strongly Connected Component | — |
| Num. Triangles | 627,584,181 |
| Num. Colors Used by LF | 86 |
| Num. Colors Used by LLF | 98 |
| Maximal Independent Set Size | 651,901 |
| Maximal Matching Size | 1,325,427 |
| Set Cover Size | 105,572 |
| (Degeneracy) | 253 |
| (Num. Peeling Rounds in -core) | 5,667 |
| Statistic | Value |
|---|---|
| Num. Vertices | 41,652,231 |
| Num. Directed Edges | 1,468,365,182 |
| Num. Undirected Edges | 2,405,026,092 |
| Effective Directed Diameter | 65 |
| Effective Undirected Diameter | 23 |
| Num. Connected Components | 2 |
| Num. Biconnected Components | 1,936,001 |
| Num. Strongly Connected Components | 8,044,729 |
| Size of Largest Connected Component | 41,652,230 |
| Size of Largest Biconnected Component | 39,708,003 |
| Size of Largest Strongly Connected Component | 33,479,734 |
| Num. Triangles | 34,824,916,864 |
| Num. Colors Used by LF | 1,081 |
| Num. Colors Used by LLF | 1,074 |
| Maximal Independent Set Size | 26,564,540 |
| Maximal Matching Size | 9,612,260 |
| Set Cover Size | 1,736,761 |
| (Degeneracy) | 2,488 |
| (Num. Peeling Rounds in -core) | 14,963 |
| Statistic | Value |
|---|---|
| Num. Vertices | 978,408,098 |
| Num. Directed Edges | 42,574,107,469 |
| Num. Undirected Edges | 74,774,358,622 |
| Effective Directed Diameter | 821 |
| Effective Undirected Diameter | 132 |
| Num. Connected Components | 23,794,336 |
| Num. Biconnected Components | 81,809,602 |
| Num. Strongly Connected Components | 135,223,661 |
| Size of Largest Connected Component | 950,577,812 |
| Size of Largest Biconnected Component | 846,117,956 |
| Size of Largest Strongly Connected Component | 774,373,029 |
| Num. Triangles | 1,995,295,290,765 |
| Num. Colors Used by LF | 4,245 |
| Num. Colors Used by LLF | 4,245 |
| Maximal Independent Set Size | 459,052,906 |
| Maximal Matching Size | 311,153,771 |
| Set Cover Size | 64,322,081 |
| (Degeneracy) | 4,244 |
| (Num. Peeling Rounds in -core) | 106,819 |
| Statistic | Value |
|---|---|
| Num. Vertices | 1,724,573,718 |
| Num. Directed Edges | 64,422,807,961 |
| Num. Undirected Edges | 124,141,874,032 |
| Effective Directed Diameter | 793 |
| Effective Undirected Diameter | 207 |
| Num. Connected Components | 129,441,050 |
| Num. Biconnected Components | 132,198,693 |
| Num. Strongly Connected Components | 1,290,550,195 |
| Size of Largest Connected Component | 1,574,786,584 |
| Size of Largest Biconnected Component | 1,435,626,698 |
| Size of Largest Strongly Connected Component | 320,754,363 |
| Num. Triangles | 4,587,563,913,535 |
| Num. Colors Used by LF | 4154 |
| Num. Colors Used by LLF | 4158 |
| Maximal Independent Set Size | 1,333,026,057 |
| Maximal Matching Size | 242,469,131 |
| Set Cover Size | 23,869,788 |
| (Degeneracy) | 4,160 |
| (Num. Peeling Rounds in -core) | 58,711 |
| Statistic | Value |
|---|---|
| Num. Vertices | 3,563,602,789 |
| Num. Directed Edges | 128,736,914,167 |
| Num. Undirected Edges | 225,840,663,232 |
| Effective Directed Diameter | 5275 |
| Effective Undirected Diameter | 331 |
| Num. Connected Components | 144,628,744 |
| Num. Biconnected Components | 298,663,966 |
| Num. Strongly Connected Components | 1,279,696,892 |
| Size of Largest Connected Component | 3,355,386,234 |
| Size of Largest Biconnected Component | 3,023,064,231 |
| Size of Largest Strongly Connected Component | 1,827,543,757 |
| Num. Triangles | 9,648,842,110,027 |
| Num. Colors Used by LF | 10,566 |
| Num. Colors Used by LLF | 10,566 |
| Maximal Independent Set Size | 1,799,823,993 |
| Maximal Matching Size | 2,434,644,438 |
| Set Cover Size | 372,668,619 |
| (Degeneracy) | 10,565 |
| (Num. Peeling Rounds in -core) | 130,728 |